/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _CLANG_CONDITION_
#define _CLANG_CONDITION_

#include <iostream>

#include "ClangTypes.h"
#include "ClangRegion.h"

class Rule;
class Condition
{
 public:
  virtual ~Condition() {}
  virtual bool Evaluate( Rule * r ) { return false; }
  virtual void Output( ostream & os, bool dv = true ) {}
} ;

struct conditionList
{
  conditionList()
  {
    cond = NULL;
    next = NULL;
  }
  
  ~conditionList()
  {
    delete cond;
  }
  
  Condition *cond;
  conditionList *next;
} ;

class TrueCondition: public Condition
{
 public:
  virtual ~TrueCondition() {}
  virtual bool Evaluate( Rule * r ) { return true; }
  virtual void Output( ostream & os, bool dv = true ) {
    os << "(true)";
  }
} ;

class FalseCondition: public Condition
{
 public:
  virtual ~FalseCondition() {}
  virtual bool Evaluate( Rule * r ) { return false; }
  virtual void Output( ostream & os, bool dv = true ) {
    os << "(false)";
  }
} ;

class PPosCondition: public Condition
{
 public:
  PPosCondition();
  virtual ~PPosCondition();
  PPosCondition( CLTeam t, UnumSet * s, int minplrs, int maxplrs,
		 Region * r );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  CLTeam t;
  UnumSet *s;
  int minplrs;
  int maxplrs;
  Region *reg;
} ;


class BPosCondition: public Condition
{
 public:
  BPosCondition();
  virtual ~BPosCondition() { delete reg; }
  BPosCondition( Region * r );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );
 private:
  Region * reg;
} ;

class BOwnerCondition: public Condition
{
 public:
  BOwnerCondition();
  virtual ~ BOwnerCondition() { delete s; }
  BOwnerCondition( CLTeam t, UnumSet * u );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  CLTeam t;
  UnumSet *s;
} ;

class PlayMCondition: public Condition
{
 public:
  PlayMCondition() {}
  virtual ~PlayMCondition() {}
  PlayMCondition( CLPlayMode pm );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  CLPlayMode pm;
} ;

enum CondCompOp { 
  lt, gt, lte, gte, eq, neq 
} ;

enum CondCompOperandType { 
  constant, svtime, ourgoals, oppgoals, goaldiff 
} ;

class CondComp: public Condition
{
 public:
  CondComp();
  virtual ~CondComp() {}
  CondComp( CondCompOperandType ot1, double const1, CondCompOp op,
	    CondCompOperandType ot2, double const2 );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  CondCompOperandType ot1;
  double const1;
  CondCompOp op;
  CondCompOperandType ot2;
  double const2;
} ;

class UnumCondition:public Condition
{
 public:
  UnumCondition();
  virtual ~ UnumCondition();
  UnumCondition( CLangVar * v, UnumSet * u );
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );
 private:
  CLangVar * v;
  UnumSet *u;
  UnumSet *n;
} ;

class AndCondition:public Condition
{
 public:
  AndCondition() 
  {
    l = NULL;
  }
  virtual ~AndCondition();
  AndCondition( conditionList * l ) 
  {
    this->l = l;
  }
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  conditionList * l;
} ;


class OrCondition: public Condition
{
 public:
  OrCondition()
  {
    l = NULL;
  }
  virtual ~OrCondition();
  OrCondition( conditionList * l ) 
  {
    this->l = l;
  }
  virtual bool Evaluate( Rule * r );
  virtual void Output( ostream & os, bool dv = true );

 private:
  conditionList * l;
} ;

class NotCondition: public Condition
{
 public:
  NotCondition()
  {
    c = NULL;
  }
  virtual ~NotCondition()
  {
    delete c;
  }
  NotCondition( Condition * c ) 
  {
    this->c = c;
  }
  virtual bool Evaluate( Rule * r )
  {
    return !c->Evaluate( r );
  }
  virtual void Output( ostream & os, bool dv = true );

 private:
  Condition * c;

} ;

class VarCondition: public Condition
{
 public:
  VarCondition( CLangVar * v )
  {
    this->v = v;
  }
  virtual ~ VarCondition()
  {
    delete v;
  }
  virtual void Output( ostream & os, bool dv = true );
  virtual bool Evaluate( Rule * r );

private:
  CLangVar * v;
} ;

#endif
