/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iomanip>
#include "ClangDirective.h"
#include "ClangRuleBase.h"
#include "ClangRule.h"

StdAction::StdAction( ActType t, Region * r, int i, UnumSet * u )
{
  type = t;
  rarg = r;
  iarg = i;
  uarg = u;
}

UnumSet *VarAction::GetUnumArg()
{
  Action *a = rulebase->ResolveAction( v );
  if ( !a )
    return NULL;
  a->SetAssocRule( r );
  return a->GetUnumArg();
}

Region *VarAction::GetRegionArg()
{
  Action *a = rulebase->ResolveAction( v );
  if ( !a )
    return NULL;
  a->SetAssocRule( r );
  return a->GetRegionArg();
}

int VarAction::GetIntArg()
{
  return rulebase->ResolveAction( v )->GetIntArg();
}

ActType VarAction::GetType()
{
  return rulebase->ResolveAction( v )->GetType();
}

/*!	Adds a player as part of the set
	\param i the player to set
*/
void UnumSet::Set( int i )
{
  if ( i == 0 ) {
    set = 0x7FF;
  }
  else {
    set |= ( 1 << ( i - 1 ) );
  }
}

/*!	Makes the UnumSet empty
*/
void UnumSet::Clear()
{
  set = 0;
}

void UnumSet::InstantiateMinus( UnumSet * s )
{
  ClearInstantiation();
  for ( int i = 1; i <= 11; i++ ) {
    if ( s->IsSet( i ) )
      Instantiate( i );
  }
}

/*!	Checks to see if a player is in the set
	\param plr the player to check
	\param checkinst if true, checks the variable's instantiation
	\retval true if the player is in the set, false otherwise
*/
bool UnumSet::IsSet( int plr, bool checkinst )
{
  if ( v == NULL ) {
    return ( ( set & ( 1 << ( plr - 1 ) ) ) == ( 1 << ( plr - 1 ) ) );
  }
  if ( checkinst ) {
    return ( ( inst & ( 1 << ( plr - 1 ) ) ) == ( 1 << ( plr - 1 ) ) );
  }
  else {
    return ( ( set & ( 1 << ( plr - 1 ) ) ) == ( 1 << ( plr - 1 ) ) );
  }
}

/*!	Removes a player from the set
	\param plr the player to remove
*/
void UnumSet::Clear( int plr )
{
  set &= ~( 1 << ( plr - 1 ) );
}

/*!	Outputs the set in its CLang parseable form
	\param os the stream to output to
	\param dispvar if false, displays the sets instantiation
*/
void UnumSet::Output( ostream & os, bool dispvar )
{
  /*cout << "(set = 0x" << hex << set << ", ";
     cout << "inst = 0x" << hex << inst << ")"; */
  int allset = 1;
  for ( int i = 1; i <= 11; i++ ) {
    if ( !IsSet( i, !dispvar ) )
      allset = 0;
  }
  if ( allset ) {
    os << "{0}";
  }
  else {
    int sp = 0;
    os << "{";
    for ( int i = 1; i <= 11; i++ ) {
      if ( IsSet( i, !dispvar ) ) {
	if ( sp ) {
	  os << ' ' << i;
	}
	else {
	  os << i;
	  sp = 1;
	}
      }
    }

    if ( dispvar && v ) {
      if ( sp )
	os << ' ';
      v->Output( os );
    }
    os << '}';
  }

}

/*!	Adds a player to the UnumSet's variable's instantiation
	\param plr The player to add
*/
void UnumSet::Instantiate( int plr )
{
  if ( v ) {
    if ( plr != -1 ) {
      if ( ( set & ( 1 << ( plr - 1 ) ) ) != ( 1 << ( plr - 1 ) ) ) {
	inst |= ( 1 << ( plr - 1 ) );
      }
    }
  }
  else {
    inst = set;
  }
}

// Action stuff

UnumSet *StdAction::GetUnumArg()
{
  if ( uarg->ContainsVariable() && r ) {
    uarg->InstantiateMinus( r->
			    GetInstantiatedVariable( uarg->
						     GetCLangVar() ) );
    return uarg;
  }
  else {
    return uarg;
  }
}

void StdAction::Output( ostream & os, bool dv )
{
  if ( type == POS ) {
    os << "(pos ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == HOME ) {
    os << "(home ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == MARK ) {
    os << "(mark ";
    GetUnumArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == MARKLREG ) {
    os << "(markl ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == MARKLUNUM ) {
    os << "(markl ";
    GetUnumArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == OLINE ) {
    os << "(oline ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == HTYPE ) {
    os << "(htype " << iarg << ')';
  }
  else if ( type == PASSREG ) {
    os << "(pass ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == PASSUNUM ) {
    os << "(pass ";
    GetUnumArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == DRIBBLE ) {
    os << "(dribble ";
    GetRegionArg()->Output( os, dv );
    os << ')';
  }
  else if ( type == CLEAR ) {
    os << "(clear ";
    GetRegionArg()->Output( os );
    os << ')';
  }
  else if ( type == SHOOT ) {
    os << "(shoot)";
  }
  else if ( type == HOLD ) {
    os << "(hold)";
  }
  else if ( type == INTERCEPT ) {
    os << "(intercept)";
  }
  else if ( type == TACKLE ) {
    os << "(tackle ";
    GetUnumArg()->Output( os, dv );
    os << ')';
  }

}

Region *StdAction::GetRegionArg()
{
  rarg->SetAssocRule( r );
  return rarg;
}

StdDirective::StdDirective( actionList * a, CLTeam t, UnumSet * s,
			    bool domode )
{
  this->a = a;
  this->t = t;
  this->s = s;
  this->domode = domode;
}

actionList *StdDirective::GetActions()
{
  actionList *c = a;
  while ( c ) {
    c->act->SetAssocRule( r );
    c = c->next;
  }
  return a;

}

UnumSet *StdDirective::GetUnum()
{
  if ( s->ContainsVariable() && r ) {
    s->InstantiateMinus( r->GetInstantiatedVariable( s->GetCLangVar() ) );
    return s;
  }
  else
    return s;
}

void StdDirective::SetAssocRule( Rule * r )
{
  this->r = r;
  actionList *c = a;
  while ( c ) {
    c->act->SetAssocRule( r );
    c = c->next;
  }

}

void StdDirective::Output( ostream & os, bool dv )
{
  if ( domode )
    os << "(do ";
  else
    os << "(dont ";
  if ( t == opp )
    os << "opp ";
  else
    os << "our ";
  GetUnum()->Output( os, dv );
  os << ' ';
  actionList *c = a;
  while ( c ) {
    c->act->Output( os, dv );
    if ( c->next )
      os << ' ';
    c = c->next;
  }
  os << ')';
}

actionList *VarDirective::GetActions()
{
  Directive *d = rulebase->ResolveDirective( v );
  if ( !d )
    return NULL;
  d->SetAssocRule( r );
  return d->GetActions();

}

CLTeam VarDirective::GetTeam()
{
  return rulebase->ResolveDirective( v )->GetTeam();
}

UnumSet *VarDirective::GetUnum()
{
  Directive *d = rulebase->ResolveDirective( v );
  if ( !d )
    return NULL;
  d->SetAssocRule( r );
  return d->GetUnum();
}

bool VarDirective::GetMode()
{
  return rulebase->ResolveDirective( v )->GetMode();
}
