/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _CLANG_DIRECTIVE_
#define _CLANG_DIRECTIVE_

#include <iostream>
#include "ClangTypes.h"
#include "ClangRegion.h"

class Rule;

typedef int ActType;
#define	POS		1
#define	HOME		2
#define	MARK		3
#define	MARKLREG	4
#define	MARKLUNUM	5
#define	OLINE		6
#define	HTYPE		7
#define	PASSREG		8
#define	PASSUNUM	9
#define	DRIBBLE		10
#define	CLEAR		11
#define	SHOOT		12
#define	HOLD		13
#define	INTERCEPT	14
#define	TACKLE		15

/// Action is the base class for the action portion of a directive
class Action
{
public:
  Action() {}
  virtual ~Action() {}

  /// Sets the type of an Action
  void SetType( ActType t ) {}

  /// Gets the type of an Action
  virtual ActType GetType()
  {
    return HTYPE;
  }
  /// Gets the argument of an action with a Unum argument
  virtual UnumSet *GetUnumArg()
  {
    return NULL;
  }
  /// Gets the argument of an action with a region argument
  virtual Region *GetRegionArg()
  {
    return NULL;
  }
  /// Gets the argument of an action with an integer argument
  virtual int GetIntArg()
  {
    return 0;
  }
  /// Sets the argument of an action with a Unum argument
  void SetUnumArg( UnumSet * s )
  {
  }
  /// Sets the argument of an action with a region argument
  void SetRegionArg( Region * r )
  {
  }
  /// Sets the argument of an action with an integer argument
  void SetIntArg( int i )
  {
  }
  /// Outputs the action to the specified ostream
  virtual void Output( ostream & os, bool dv = true ) {}

  virtual void SetAssocRule( Rule * r ) {}

  virtual Rule *GetAssocRule()
  {
    return NULL;
  }
} ;

/// A StdAction is an action class representing all action types except for an action referencing a predefined action variable
class StdAction: public Action
{
public:
  StdAction()
  {
    rarg = NULL;
    uarg = NULL;
    r = NULL;
  }
  /// Creates a StdAction of the specified type with the specified arguments
  StdAction( ActType t, Region * r, int i, UnumSet * u );
  virtual ~StdAction() {
    delete rarg;
    delete uarg;
  }
  void SetType( ActType t )
  {
    type = t;
  }

  virtual ActType GetType()
  {
    return type;
  }
  virtual UnumSet *GetUnumArg();
  virtual Region *GetRegionArg();
  virtual int GetIntArg()
  {
    return iarg;
  }

  void SetUnumArg( UnumSet * s )
  {
    uarg = s;
  }
  void SetRegionArg( Region * r )
  {
    rarg = r;
  }
  void SetIntArg( int i )
  {
    iarg = i;
  }
  virtual void Output( ostream & os, bool dv = true );
  virtual void SetAssocRule( Rule * r )
  {
    this->r = r;
  }
  virtual Rule *GetAssocRule( Rule * r )
  {
    return r;
  }

private:
  ActType type;
  Region *rarg;
  int iarg;
  UnumSet *uarg;
  Rule *r;
} ;

/// A VarAction is a special action class representing actions of the form "<actionvar>" (i.e. those referencing a predefined action variable

class VarAction: public Action
{
public:
  /// Creates a VarAction that references the given variable
  VarAction( CLangVar * v )
  {
    this->v = v;
    r = NULL;
  }

  ~VarAction()
  {
    delete v;
  }

  virtual ActType GetType();
  virtual UnumSet *GetUnumArg();
  virtual Region *GetRegionArg();
  virtual int GetIntArg();
  virtual void Output( ostream & os, bool dv = false ) 
  {
    v->Output( os );
  }
  virtual void SetAssocRule( Rule * r )
  {
    this->r = r;
  }
  virtual Rule *GetAssocRule()
  {
    return r;
  }
private:
  CLangVar * v;
  Rule *r;
} ;

struct actionList
{
  actionList()
  {
    act = NULL;
    next = NULL;
  }

  ~actionList()
  {
    delete act;
  }
  Action *act;
  actionList *next;
} ;

class Directive
{
public:
  virtual ~Directive() {}

  virtual void Output( ostream & os, bool dv = true ) {}

  virtual actionList *GetActions()
  {
    return NULL;
  }
  virtual CLTeam GetTeam()
  {
    return our;
  }
  virtual UnumSet *GetUnum()
  {
    return NULL;
  }
  virtual bool GetMode()
  {
    return true;
  }
  virtual void SetAssocRule( Rule * r ) {}

  virtual Rule *GetAssocRule()
  {
    return NULL;
  }
} ;

struct directiveList
{
  directiveList()
  {
    dir = NULL;
    next = NULL;
  }
   ~directiveList()
  {
    delete dir;
  }
  Directive *dir;
  directiveList *next;
} ;

class StdDirective: public Directive
{
public:
  StdDirective()
  {
    a = NULL;
    s = NULL;
    r = NULL;
  }
  virtual ~StdDirective()
  {
    delete s;
    actionList *c = a, *t;
    while ( c ) {
      t = c;
      c = c->next;
      delete t;
    }
  }

  StdDirective( actionList * a, CLTeam t, UnumSet * s, bool domode );
  virtual void Output( ostream & os, bool dv = true );
  virtual actionList *GetActions();
  virtual CLTeam GetTeam()
  {
    return t;
  }
  virtual UnumSet *GetUnum();
  virtual bool GetMode()
  {
    return domode;
  }
  virtual void SetAssocRule( Rule * r );
  virtual Rule *GetAssocRule()
  {
    return r;
  }

private:
  actionList * a;
  CLTeam t;
  UnumSet *s;
  bool domode;
  Rule *r;
} ;

class VarDirective: public Directive
{
public:
  VarDirective( CLangVar * v )
  {
    this->v = v;
    r = NULL;
  }
  virtual ~ VarDirective()
  {
    delete v;
  }
  virtual void Output( ostream & os, bool dv = true ) {
    v->Output( os );
  }
  virtual actionList *GetActions();
  virtual CLTeam GetTeam();
  virtual UnumSet *GetUnum();
  virtual bool GetMode();
  virtual void SetAssocRule( Rule * r )
  {
    this->r = r;
  }
  virtual Rule *GetAssocRule()
  {
    return r;
  }

private:
  CLangVar * v;
  Rule *r;
} ;

#endif
