/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "ClangEval.h"
#include "ClangRBInterface.h"
#include <strstream>
#include <fstream>

using namespace std;

CLEval::CLEval( WorldModel *wm, ServerSettings *ss, PlayerSettings *ps, char *inFile )
{
  this->wm = wm;
  this->ss = ss;
  this->ps = ps;
  evals = NULL;
  stockactions = NULL;
  numStockActions = 0;

  // stuff from the stockactions.conf
  passanglerange = 10;
  passanglestep = 5;

  passspeedmin = 1;
  passspeedmax = 2.5;
  passspeedstep = .5;

  passvalue = 14;
  passdelta = 4.5;

  // 

  shotvalue = 12;
  shotstep = 1;
  shotdelta = 2;
  clearvalue = 4.6;
  dribblevalue = 8;
  fastdribblethr = 1;
  fastdribblethr = 4;
  dribblerad = 9;
  dribbledelta = 2;

  minpassconf = 0;

  stockclrlx = 0;
  stockclrly = -30;
  stockclrrx = 0;
  stockclrry = 30;
  stockclrcx = 40;
  stockclrcy = 0;

  stockdriblx = 40;
  stockdribly = -20;
  stockdribrx = 40;
  stockdribry = 20;
  stockdribcx = 35;
  stockdribcy = 0;

  if ( inFile != NULL )
    LoadOptions( inFile );
}

CLEval::~CLEval()
{
  Cleanup();
  for ( int i = 0; i < numStockActions; i++ ) {
    delete stockactions[i];
  }

  delete[]stockactions;

}

void CLEval::Cleanup()
{
  CLEvalList *cur = evals, *tmp;
  while ( cur ) {
    tmp = cur;
    cur = cur->next;
    delete tmp;
  }
  evals = NULL;

}

void CLEval::Update()
{
  Cleanup();

  for ( int i = 0; i < 11; i++ )
    passadvised[i] = NO_ADVICE;

  shootadvised = NO_ADVICE;
  holdadvised = NO_ADVICE;
  interceptadvised = NO_ADVICE;

  passtoreceive = false;
  int mynum = wm->getPlayerNumber();

  // Reset home Position to something similar
  // to TriWimpy's 433_OFFENSIVE
  // This is an ultimate last resort
  // This position will only be used if a GO_HOME is tried and
  // there were no home or pos directives
  double hx[11] = { -50, -16, -17, -17, -16, -8, -5, -5, -2, -1, -1 };
  double hy[11] = { 0, 16, 5, -5, -16, 0, 10, -10, 0, 22, -22 };
  homepos.setVecPosition( hx[mynum - 1], hy[mynum - 1] );

  usepos = false;

  Log.log( 650, "Applicable Directives:" );

  RBlist *dl = rulebase->GetActiveDirecs();

  while ( dl ) {
    if ( dl->dir->GetUnum()->IsSet( mynum ) ) {

      strstream ostr;
      ostr << "   ";
      dl->dir->Output( ostr, false );
      ostr << ends;
      Log.log( 650, ostr.str() );
      actionList *al = dl->dir->GetActions();
      while ( al ) {
	if ( al->act->GetType() == PASSUNUM
	     || al->act->GetType() == PASSREG ) {
	  UpdatePassAdvice( al->act, dl->dir->GetMode() );
	  Log.log( 1000, "END PASSADVICE" );
	}
	else if ( al->act->GetType() == HOME ) {
	  Region *reg = al->act->GetRegionArg();
	  UpdateHomePosition( reg );
	}
	else if ( al->act->GetType() == POS ) {
	  Region *reg = al->act->GetRegionArg();
	  UpdatePosPosition( reg );
	  usepos = true;

	}
	else if ( al->act->GetType() == HTYPE ) {
	}
	else if ( al->act->GetType() == MARK
		  || al->act->GetType() == MARKLUNUM
		  || al->act->GetType() == MARKLREG ) {
	}
	else if ( al->act->GetType() == SHOOT ) {
	  if ( dl->dir->GetMode() ) {
	    shootadvised = ADVISED_DO;
	  }
	  else {
	    shootadvised = ADVISED_DONT;
	  }
	}
	else if ( al->act->GetType() == INTERCEPT ) {
	  if ( dl->dir->GetMode() ) {
	    interceptadvised = ADVISED_DO;
	  }
	  else {
	    interceptadvised = ADVISED_DONT;
	  }
	}
	else if ( al->act->GetType() == HOLD ) {
	  if ( dl->dir->GetMode() ) {
	    holdadvised = ADVISED_DO;
	  }
	  else {
	    holdadvised = ADVISED_DONT;
	  }
	}

	else if ( dl->dir->GetMode() ) {
	  CLEvalList *ev = new CLEvalList;
	  strstream tmpstr;
	  al->act->Output( tmpstr, false );
	  tmpstr << ends;
	  ev->act = al->act;
	  ev->next = evals;
	  evals = ev;
	}
	al = al->next;
      }

    }

    dl = dl->next;
  }

  for ( int i = 0; i < numStockActions; i++ ) {

    // Never consider a pass to yourself (of course)
    if ( stockactions[i]->GetType() != STOCK_PASS
	 || stockactions[i]->GetIntArg() != mynum ) {
      CLEvalList *stnew = new CLEvalList;
      stnew->act = stockactions[i];
      stnew->next = evals;
      evals = stnew;
    }
  }

  Log.log( 699, "Choosing shot point" );
  ChooseShootPt();
  Log.log( 699, "Choosing pass points" );
  ChoosePasses();
  Log.log( 699, "Choosing marks" );
  ChooseMarks();
  Log.log( 699, "Done" );
}

void CLEval::UpdatePassAdvice( Action * a, bool doordont )
{
  if ( a->GetType() == PASSUNUM ) {
    UnumSet *us = a->GetUnumArg();
    for ( int i = 0; i < 11; i++ ) {
      if ( us->IsSet( i + 1 ) ) {
	if ( doordont ) {
	  Log.log( 1000, "DO %d", i );
	  passadvised[i] = ADVISED_DO;
	}
	else {
	  Log.log( 1000, "DONT %d", i );
	  passadvised[i] = ADVISED_DONT;
	}
      }
    }
  }
  else if ( a->GetType() == PASSREG ) {
    Region *reg = a->GetRegionArg();
    VecPosition p;
    for ( int i = 0; i < 11; i++ ) {
      p =
	wm->getGlobalPosition( SoccerTypes::getOpponentObjectFromIndex( i ) );

      if ( reg->PointInRegion( p.getX(), p.getY() ) ) {
	if ( doordont ) {
	  Log.log( 1000, "DO %d", i );
	  passadvised[i] = ADVISED_DO;
	}
	else {
	  Log.log( 1000, "DONT %d", i );
	  passadvised[i] = ADVISED_DONT;
	}
      }

    }
  }

}

Action *CLEval::GetBestAction()
{
  CLEvalList *c = evals;
  Action *best = NULL;
  double bestscore = -1;
  Log.log( 651, "Evaluating actions:" );
  while ( c ) {

    c->eval = Evaluate( c->act );
    if ( c->eval > bestscore ) {
      best = c->act;
      bestscore = c->eval;
    }
    strstream ostr;
    ostr << "   (%d) ";
    c->act->Output( ostr, false );
    ostr << ": %f";
    ostr << ends;
    Log.log( 651, ostr.str(), c->act->GetType(), c->eval );
    c = c->next;
  }

  return best;
}

double CLEval::Evaluate( Action * a )
{
  PointRegion *tmp;
  double x;
  switch ( a->GetType() ) {
  case STOCK_PASS:
    if ( wm->isBallKickable() ) {
      int pto = a->GetIntArg();
      x = EvalPass( pto );
      if ( passadvised[pto - 1] == NO_ADVICE )
	return x;
      else if ( passadvised[pto - 1] == ADVISED_DO )
	return x * doweight;
      else if ( passadvised[pto - 1] == ADVISED_DONT )
	return x * dontweight;

    }
    else {
      return 0;
    }
    break;
  case STOCK_DRIBBLE_LEFT:
    tmp = new PointRegion( stockdriblx, stockdribly );
    x = EvalDribble( tmp );
    delete tmp;
    if ( DribbleUnadvised( stockdriblx, stockdribly ) )
      x *= dontweight;
    return x;
    break;
  case STOCK_DRIBBLE_RIGHT:
    tmp = new PointRegion( stockdribrx, stockdribry );
    x = EvalDribble( tmp );
    delete tmp;
    if ( DribbleUnadvised( stockdribrx, stockdribry ) )
      x *= dontweight;

    return x;
    break;
  case STOCK_DRIBBLE_CENTER:
    tmp = new PointRegion( stockdribcx, stockdribcy );
    x = EvalDribble( tmp );
    delete tmp;
    if ( DribbleUnadvised( stockdribcx, stockdribcy ) )
      x *= dontweight;

    return x;
    break;
  case STOCK_CLEAR_LEFT:
    tmp = new PointRegion( stockclrlx, stockclrly );
    x = EvalClear( tmp );
    if ( wm->getBallPos().getX() < 0 )
      x *= 1.2;
    delete tmp;
    if ( ClearUnadvised( stockclrlx, stockclrly ) )
      x *= dontweight;
    return x;

    break;
  case STOCK_CLEAR_RIGHT:
    tmp = new PointRegion( stockclrrx, stockclrry );
    x = EvalClear( tmp );
    if ( wm->getBallPos().getX() < 0 )
      x *= 1.2;
    delete tmp;
    if ( ClearUnadvised( stockclrrx, stockclrry ) )
      x *= dontweight;

    return x;

    break;
  case STOCK_CLEAR_CENTER:
    tmp = new PointRegion( stockclrcx, stockclrcy );
    x = EvalClear( tmp );
    delete tmp;

    // Really shouldn't center it when far behind the center point, because
    // then a "center" is really a shot
    if ( wm->getBallPos().getX() < 0 )
      x *= 2.5;

    if ( wm->getBallPos().getX() > stockclrcx )
      return x;
    else
      return x * lessthandelta( wm->getBallPos().getX(), stockclrcx, 10 );

    break;
  case STOCK_GO_HOME:
    if ( wm->isBallKickable() )
      return 0;
    if ( usepos )
      return 0;
    else
      return gohomeconst;

    break;
  case STOCK_GO_POS:
    if ( wm->isBallKickable() ) {
      Log.log( 799, "Not intercepting because ball is kickable" );
      return 0;
    }
    if ( usepos )
      return gohomeconst;
    else
      return 0;

    break;
  case STOCK_HOLD:
    if ( wm->isBallKickable() ) {
      if ( holdadvised == NO_ADVICE )
	return holdconst;
      else if ( holdadvised == ADVISED_DO )
	return holdconst * doweight;
      else if ( holdadvised == ADVISED_DONT )
	return holdconst * dontweight;
    }
    else
      return 0;
    break;
  case STOCK_INTERCEPT:
    if ( wm->isBallKickable() ) 
      return 0;
    //else if (SoccerTypes::isTeammate(wm->getBallOwner()))
    //{
    //        return 0;
    //}
    else {
      x = EvalIntercept();
      if ( interceptadvised == NO_ADVICE )
	return x;
      else if ( interceptadvised == ADVISED_DO )
	return x * doweight;
      else if ( interceptadvised == ADVISED_DONT )
	return x * dontweight;

    }
    break;
  case STOCK_GOALIE_CATCH:
    if ( wm->getPlayerNumber() == 1 && wm->isBallCatchable() )
      return 10000;
    else
      return 0;
    break;
  case STOCK_SHOOT:
    if ( shootadvised == NO_ADVICE )
      return EvalShoot();
    else if ( shootadvised == ADVISED_DO )
      return EvalShoot() * doweight;
    else if ( shootadvised == ADVISED_DONT )
      return EvalShoot() * dontweight;
    break;
  case DRIBBLE:
    return EvalDribble( a->GetRegionArg() ) * doweight;
    break;
  case CLEAR:
    return EvalClear( a->GetRegionArg() ) * doweight;
    break;
  case STOCK_MARK:
    return markeval;
    break;
  case STOCK_RECEIVE_PASS:
    if ( passtoreceive ) {
      return interceptclosest + 1;
    }
    else {
      return 0;
    }
    break;
  case TACKLE:
    return 0;
    break;
  }
  cout << "Unknown action... type: " << a->GetType() << " iarg: " << a->
    GetIntArg() << endl;
  return -99;
}

double CLEval::EvalIntercept()
{
  if ( wm->isDeadBallThem() || wm->isKickOffThem()
       || wm->getPlayMode() == PM_BEFORE_KICK_OFF || wm->isKickInThem()
       || wm->isFreeKickThem() || wm->isGoalKickThem() ) {
    Log.log( 799, "Not evaluating intercept %s %d %d\n", 
	     SoccerTypes::getPlayModeStr( wm->getPlayMode() ), 
	     wm->getSide(), wm->isKickOffThem() );
    if ( !wm->isKickOffUs() )
    return 0;
  }
  char buffer[ 100 ];
  SoccerTypes::getObjectStr( buffer, wm->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL ) );
  if ( wm->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL ) ==
       wm->getAgentObjectType() )
    return interceptfastest;
  else if ( wm->getClosestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL ) ==
	    wm->getAgentObjectType() )
    return interceptclosest;
  else
    return interceptnotfastest;
}

double CLEval::EvalPass( int passto )
{

  if ( passto == 1 )
    return 0;
  return passeval[passto - 1];

}

void CLEval::UpdateMarkAdvice( Directive * d )
{
  actionList *al = d->GetActions();
  UnumSet *dset = d->GetUnum();
  UnumSet *aset;
  while ( al ) {
    Action *a = al->act;

    if ( a->GetType() == MARK ) {
      aset = a->GetUnumArg();
      for ( int i = 0; i < 11; i++ ) {
	for ( int j = 0; j < 11; j++ ) {
	  if ( dset->IsSet( i + 1 ) && aset->IsSet( j + 1 ) ) {
	    markmatrix[i][j] = true;
	    Log.log( 705, "Adding %d,%d to the Mark Matrix", i + 1, j + 1 );
	  }
	}
      }
    }
    else if ( a->GetType() == MARKLREG ) {
      Region *r = a->GetRegionArg();
      VecPosition opos;
      for ( int i = 0; i < 11; i++ ) {
	for ( int j = 0; j < 11; j++ ) {
	  opos =
	    wm->
	    getGlobalPosition( SoccerTypes::getOpponentObjectFromIndex( j ) );
	  if ( dset->IsSet( i + 1 )
	       && r->PointInRegion( opos.getX(), opos.getY() ) ) {
	    marklmatrix[i][j] = true;

	    Log.log( 705, "Adding %d,%d to the MarkL Matrix", i + 1, j + 1 );
	  }

	}
      }
    }
    else if ( a->GetType() == MARKLUNUM ) {
      aset = a->GetUnumArg();
      for ( int i = 0; i < 11; i++ ) {
	for ( int j = 0; j < 11; j++ ) {
	  if ( dset->IsSet( i + 1 ) && aset->IsSet( j + 1 ) ) {
	    markmatrix[i][j] = true;
	    Log.log( 705, "Adding %d,%d to the MarkL Matrix", i + 1, j + 1 );
	  }
	}
      }

    }
    al = al->next;
  }
}

double CLEval::EvalShoot()
{
  if ( !wm->isBallKickable() )
    return 0;

  return shoteval;
}

double CLEval::EvalDribble( Region * r )
{
  if ( wm->getPlayerNumber() == 1 )
    return 0;
  if ( wm->getBallOwner() != wm->getAgentObjectType()
       && !wm->isBallKickable() )
    return 0;
  if ( wm->isDeadBallUs() )	// Can't dribble on kick offs
    return 0;

  VecPosition dribvel, ballpos = wm->getBallPos();
  double tx, ty;
  r->GetClosestPointInRegion( ballpos.getX(), ballpos.getY(), &tx, &ty );

  if ( ballpos.getDistanceTo( VecPosition( tx, ty ) ) < 2 )
    return 0;

  double kickdist = 4;

  dribvel =
    ( VecPosition( tx, ty ) - ballpos ).normalize() * kickdist / 2.5 * 10;

  ObjectT opp;
  int lotime = 9999, opptime;
  for ( int i = 0; i < 11; i++ ) {
    opp = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
      opptime = wm->CyclesToIntercept( opp, dribvel );
      if ( opptime < lotime )
	lotime = opptime;
    }
  }

  if ( lotime < fastdribblethr )
    kickdist = 10;
  else if ( lotime < slowdribblethr )
    kickdist = 5;
  else
    kickdist = 4;

  dribvel =
    ( VecPosition( tx, ty ) - ballpos ).normalize() * kickdist / 2.5 * 10;
  int mytime =
    wm->
    CyclesToIntercept( SoccerTypes::
		       getTeammateObjectFromIndex( wm->
						   getPlayerNumber() -
						   1 ), dribvel );

  lotime = 9999;
  for ( int i = 0; i < 11; i++ ) {
    opp = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
      opptime = wm->CyclesToIntercept( opp, dribvel );
      if ( opptime < lotime )
	lotime = opptime;
    }
  }
  double prob = lessthandelta( mytime, lotime, dribbledelta );
  Log.log( 699, "Mytime: %d, lotime: %d, prob %.4f", mytime, lotime, prob );
  return prob * dribblevalue;
}

double CLEval::EvalClear( Region * r )
{
  if ( !wm->isBallKickable() )
    return 0;

  VecPosition clrpt, ballpos = wm->getBallPos(), ip, qip;
  int qitime = 9999, itime;
  double tx, ty;
  r->GetClosestPointInRegion( ballpos.getX(), ballpos.getY(), &tx, &ty );
  clrpt.setVecPosition( tx, ty );
  if ( clrpt.getDistanceTo( ballpos ) < cleardist )
    return 0;

  VecPosition vel =
    ( clrpt - ballpos ).normalize() * ss->getBallSpeedMax() * 10;
  ObjectT opp;
  for ( int o = 0; o < 11; o++ ) {
    opp = SoccerTypes::getOpponentObjectFromIndex( o );
    if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
      itime = wm->CyclesToIntercept( opp, vel, &ip );
      if ( itime < qitime ) {
	qitime = itime;
	qip = ip;
      }
    }
  }

  double benefit = 0;

  float x_pos_comp = 30 * ( ( qip.getX() + 52.5 ) / 105 );
  x_pos_comp += 20 * lessthandelta( 40, qip.getX(), 10 );

  float corner_dist = qip.getDistanceTo( VecPosition( 52.5, -34 ) );

  if ( qip.getDistanceTo( VecPosition( 52.5, 34 ) ) < corner_dist )
    corner_dist = qip.getDistanceTo( VecPosition( 52.5, 34 ) );

  x_pos_comp =
    ifstar( lessthandelta( corner_dist, 30, 20 ), 100, x_pos_comp );

  float goal_dist = qip.getDistanceTo( VecPosition( 52.5, 0 ) );
  benefit =
    ifstar( lessthandelta( goal_dist, shootdist, 5 ), 100, x_pos_comp ) / 100;

  // if (qip.getX() > ballpos.getX())
  //{
  //      benefit = (qip.getX() - ballpos.getX()) / (52.5 - ballpos.getX()) * (1-forwardpassbias) + forwardpassbias;
  //}
  if ( qip.getX() < ballpos.getX() )
    benefit *= .3;
  return benefit * clearvalue;
  //return qip.getDistanceTo(ballpos) / ip.getDistanceTo(ballpos) * clearvalue;

}

void CLEval::LoadOptions( char *file )
{
  ifstream opt( file );

  char line[255];
  char *colon, *data;
  int curact = 0;

  int noptions = 35;

  char *optstr[] = {
    "do_weight",
    "dont_weight",
    "gohome_value",
    "intercept_value_fastest",
    "intercept_value_closest",
    "intercept_value_other",
    "hold_value",
    "pass_angle_range",
    "pass_angle_step",
    "pass_speed_min",
    "pass_speed_max",
    "pass_speed_step",
    "pass_value",
    "pass_delta",
    "shot_value",
    "shot_delta",
    "shot_min",
    "shot_max",
    "shot_step",
    "clear_value", "dribble_value", "fast_dribble_thr", "slow_dribble_thr",
    "dribble_delta", "mark_charge_dist", "ball_attr_x", "ball_attr_y",
    "ball_attr_dist", "opp_attr_x", "opp_attr_y", "opp_attr_dist",
    "forward_pass_bias", "clear_dist", "shoot_dist"
    ,"min_pass_conf"
  };

  double *value[] = { &doweight, &dontweight, &gohomeconst, &interceptfastest,
    &interceptclosest, &interceptnotfastest, &holdconst, &passanglerange,
    &passanglestep,
    &passspeedmin, &passspeedmax, &passspeedstep, &passvalue, &passdelta,
    &shotvalue,
    &shotdelta, &shotmin, &shotmax, &shotstep, &clearvalue, &dribblevalue,
    &fastdribblethr,
    &slowdribblethr, &dribbledelta, &markchargedist, &ballattrx, &ballattry,
    &ballattrdist,
    &oppattrx, &oppattry, &oppattrdist, &forwardpassbias, &cleardist,
    &shootdist
		      ,&minpassconf
  };

  while ( !opt.eof() ) {
    opt.getline( line, 255 );
    colon = strstr( line, ": " );
    if ( colon ) {
      data = strrchr( colon, ' ' ) + 1;
      *colon = '\0';

      if ( strcmp( line, "numstockactions" ) == 0 ) {
	numStockActions = atoi( data );
	stockactions = new Action *[numStockActions];
      }
      else if ( strcmp( line, "stockaction" ) == 0 ) {
	if ( data[0] == '(' ) {
	  char *cmsg = new char[8192];
	  sprintf( cmsg, "(define (definea \"$stock%d\" %s))", curact, data );
	  rulebase->HearMsg( cmsg );
	  sprintf( cmsg, "\"$stock%d\"", curact );
	  CLangVar v( cmsg );
	  stockactions[curact++] = rulebase->ResolveAction( &v );
	  delete[]cmsg;
	}
	else {

	  if ( strncmp( data, "STOCK_PASS", 10 ) == 0 ) {
	    stockactions[curact++] =
	      new StockAction( STOCK_PASS, atoi( &data[11] ) );
	  }
	  else if ( strcmp( data, "STOCK_SHOOT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_SHOOT );
	  }
	  else if ( strcmp( data, "STOCK_INTERCEPT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_INTERCEPT );
	  }
	  else if ( strcmp( data, "STOCK_GO_HOME" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_GO_HOME );
	  }
	  else if ( strcmp( data, "STOCK_GO_POS" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_GO_POS );
	  }
	  else if ( strcmp( data, "STOCK_CLEAR_LEFT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_CLEAR_LEFT );
	  }
	  else if ( strcmp( data, "STOCK_CLEAR_RIGHT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_CLEAR_RIGHT );
	  }
	  else if ( strcmp( data, "STOCK_CLEAR_CENTER" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_CLEAR_CENTER );
	  }
	  else if ( strcmp( data, "STOCK_DRIBBLE_LEFT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_DRIBBLE_LEFT );
	  }
	  else if ( strcmp( data, "STOCK_DRIBBLE_RIGHT" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_DRIBBLE_RIGHT );
	  }
	  else if ( strcmp( data, "STOCK_DRIBBLE_CENTER" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_DRIBBLE_CENTER );
	  }
	  else if ( strcmp( data, "STOCK_GOALIE_CATCH" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_GOALIE_CATCH );
	  }
	  else if ( strcmp( data, "STOCK_HOLD" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_HOLD );
	  }
	  else if ( strcmp( data, "STOCK_RECEIVE_PASS" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_RECEIVE_PASS );
	  }
	  else if ( strcmp( data, "STOCK_MARK" ) == 0 ) {
	    stockactions[curact++] = new StockAction( STOCK_MARK );
	  }

	}
      }
      else {
	for ( int i = 0; i < noptions; i++ ) {
	  if ( strcmp( optstr[i], line ) == 0 )
	    *( value[i] ) = atof( data );
	}
      }
    }

  }

  //cout << "hold_value: " << holdconst << endl;
  //cout << "pass_value: " << passvalue << endl;
  //cout << "shot_value: " << shotvalue << endl;
  //cout << "clear_value: " << clearvalue << endl;
  //cout << "dribble_value: " << dribblevalue << endl;

  opt.close();

}

void CLEval::ChooseShootPt()
{
  ObjectT opp;

  ObjectT lo;
  VecPosition lop;
  VecPosition shot;
  VecPosition vel;
  VecPosition ip;
  VecPosition ballpos = wm->getBallPos();

  VecPosition retshot;
  double bestscore = -1;
  double score;
  double leasttime = 9999, testtime, cyclestogoal;
  // PointRegion gc(52.5,0);
  //ArcRegion tmp(&gc,0,shootdist,90,270);
  //Log.logRegion(700,&tmp,ColorI(0,0,255));

  //Log.log(698," Shoot distance = %.2f",shootdist);

  if ( !wm->isBallKickable() ||
       wm->getRelativeDistance(SoccerTypes::getGoalOpponent(wm->getSide())) > 20 ) {
    shoteval = 0;
    return;
  }
  for ( double y = shotmin; y < shotmax; y += shotstep ) {
    shot.setVecPosition( 52.5, y );
    Log.log( 698, "Checking shot on (%.2f, %.2f): distance = %.2f",
	     shot.getX(), shot.getY(), shot.getDistanceTo( ballpos ) );
    leasttime = 9999;
    for ( int o = 0; o < 11; o++ ) {
      opp = SoccerTypes::getOpponentObjectFromIndex( o );
      vel =
	( shot -
	  ballpos ).normalize() * ( ss->getBallSpeedMax() ) * 10;
      if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
	testtime = wm->CyclesToIntercept( opp, vel, &ip );
	if ( ip.getX() < 52.5 && testtime < leasttime ) {
	  lo = opp;
	  lop = ip;
	  leasttime = testtime;

	}

      }
    }

    cyclestogoal =
      shot.getDistanceTo( ballpos ) / ss->getBallSpeedMax();

    score = lessthandelta( cyclestogoal, leasttime, shotdelta );
    if ( shot.getDistanceTo( ballpos ) > shootdist ) {
      score *= lessthandelta( shootdist, shot.getDistanceTo( ballpos ), 10 );
    }

    if ( wm->getConfidence( wm->getOppGoalieType() ) <
	 ps->getOtherPlayerConfThr() )
      score *= .6;

    Log.log( 698,
	     "For shot on (%.2f, %.2f): Cycles to goal = %.2f, Least opponent time = %.2f, Score = %.2f",
	     shot.getX(), shot.getY(), cyclestogoal, leasttime, score );

    //if (shot.getDistanceTo(ballpos) > shootdist)
    //{
    //      score *= 1 - ( shot.getDistanceTo(ballpos)-shootdist ) / (shot.getDistanceTo(VecPosition(0,0)) -shootdist );

    //}

    if ( score > bestscore ) {
      bestscore = score;
      retshot = shot;
    }
//     Log.logPoint( 698, lop, ColorI( 255, 0, 0 ) );
//     Log.logLine( 698, wm->getGlobalPosition( lo ), lop, ColorI( 255, 0, 0 ) );
//     Log.logLine( 698, wm->getBallPos(), shot, ColorI( 0, 0, 255 ) );

  }
  shotpt = retshot;
  shoteval = bestscore * shotvalue;

}

void CLEval::ChoosePasses()
{
  ObjectT tm, opp, lopp = OBJECT_ILLEGAL, 
    bestlopp = OBJECT_ILLEGAL;
  double theta;
  VecPosition bestpoint;
  VecPosition tp;
  double bestspeed = 0;
  double bestprob = -1;
  double prob;
  int tmtime;
  int besttmtime = 9999;
  int opptime = 0;
  int lotime = 9999, bestlotime = 0;

  for ( int i = 0; i < 11; i++ ) {
    passeval[i] = 0;
  }

  if ( !wm->isBallKickable() )
    return;

  VecPosition tmpos;
  VecPosition mypos = wm->getAgentGlobalPosition();
  VecPosition pt, lop, bestlop;
  for ( int i = 0; i < 11; i++ ) {
    // Don't pass to self or goalie
    if ( i != wm->getPlayerNumber() - 1
	 && i != SoccerTypes::getIndex( wm->getOwnGoalieType() ) ) {
      bestprob = -1;
      tm = SoccerTypes::getTeammateObjectFromIndex( i );
      if ( wm->getConfidence( tm ) > ps->getOtherPlayerConfThr() ) {
	tmpos = wm->getGlobalPosition( tm );
	theta =
	  atan2( tmpos.getY() - mypos.getY(),
		 tmpos.getX() - mypos.getX() ) * 180 / 3.1415926;
	for ( double th = theta - passanglerange;
	      th < theta + passanglerange; th += passanglestep ) {
	  for ( double speed = passspeedmin; speed < passspeedmax;
		speed += passspeedstep ) {

	    tmtime =
	      wm->CyclesToIntercept( tm,
				     VecPosition( cos
						  ( th *
						    3.1415926 /
						    180 ) *
						  speed * 10,
						  sin( th *
						       3.1415926
						       / 180 ) *
						  speed * 10 ), &tp );
	    if ( tp.getX() > -52.5 && tp.getX() < 52.5
		 && tp.getY() > -34 && tp.getY() < 34 ) {
	      lotime = 99999;
	      for ( int o = 0; o < 11; o++ ) {
		opp = SoccerTypes::getOpponentObjectFromIndex( o );
		if ( wm->getConfidence( opp ) >
		     ps->getOtherPlayerConfThr() ) {
		  opptime =
		    wm->CyclesToIntercept( opp,
					   VecPosition( cos
							( th
							  * 3.1415926 / 180 )
							*
							speed
							*
							10,
							sin
							( th
							  * 3.1415926 / 180 )
							* speed * 10 ), &pt );

		  if ( opptime < lotime ) {
		    lotime = opptime;
		    lop = pt;
		    lopp = opp;
		  }

		}

	      }

	      double benefit = 0;
	      float x_pos_comp = 30 * ( ( pt.getX() + 52.5 ) / 105 );
	      x_pos_comp += 20 * lessthandelta( 40, pt.getX(), 10 );

	      float corner_dist =
		pt.getDistanceTo( VecPosition( 52.5, -34 ) );

	      if ( pt.getDistanceTo( VecPosition( 52.5, 34 ) ) < corner_dist )
		corner_dist = pt.getDistanceTo( VecPosition( 52.5, 34 ) );

	      x_pos_comp =
		ifstar( lessthandelta( corner_dist, 30, 20 ),
			100, x_pos_comp );

	      //return x_pos_comp;

	      float goal_dist = pt.getDistanceTo( VecPosition( 52.5, 0 ) );
	      //x_pos_comp should be less than value of being near the goal
	      benefit =
		ifstar( lessthandelta
			( goal_dist, shootdist, 5 ), 100, x_pos_comp ) / 100;

	      prob =
		lessthandelta( tmtime, lotime,
			       passdelta ) * benefit * passvalue;

	      if ( prob > bestprob ) {
		bestpoint = tp;
		bestlop = lop;
		bestprob = prob;
		bestspeed = speed;
		besttmtime = tmtime;
		bestlopp = lopp;
		bestlotime = lotime;
	      }
	    }
	  }
	}

// 	Log.logPoint( 697, bestpoint, ColorI( 0, 0, 255 ) );
// 	Log.logPoint( 697, bestlop, ColorI( 255, 0, 0 ) );
// 	Log.logLine( 697, mypos, bestpoint, ColorI( 0, 255, 0 ) );
// 	Log.logLine( 697, wm->getGlobalPosition( tm ), bestpoint,
// 		     ColorI( 0, 0, 255 ) );
// 	Log.logLine( 697, bestlop, wm->getGlobalPosition( bestlopp ),
// 		     ColorI( 255, 0, 0 ) );
	Log.log( 697,
		 "Pass selected for player %d is at (%.2f, %.2f) speed = %.3f itime = %d (can be intercepted by opponent %d at %.2f, %.2f in %d cycles): score = %.3f",
		 i + 1, bestpoint.getX(), bestpoint.getY(),
		 bestspeed, besttmtime,
		 SoccerTypes::getIndex( bestlopp ) + 1,
		 bestlop.getX(), bestlop.getY(), bestlotime, bestprob );
	passpt[i] = bestpoint;
	passeval[i] = bestprob;
	passspd[i] = bestspeed;
	passtime[i] = besttmtime;
	passtargconf[i] = wm->getConfidence(tm);
      }
    }
  }

}

void CLEval::UpdateHomePosition( Region * r )
{
  double px, py;
  VecPosition mypos = wm->getAgentGlobalPosition();
  VecPosition ballpos, opppos;
  ObjectT opp;

  r->GetClosestPointInRegion( ballpos.getX(), ballpos.getY(), &px, &py );
  VecPosition p( px, py );

  if ( p.getDistanceTo( ballpos ) <= ballattrdist ) {
    px += ballpos.getX() * ballattrx;
    py += ballpos.getY() * ballattry;
  }

  for ( int i = 0; i < 11; i++ ) {
    opp = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
      opppos = wm->getGlobalPosition( opp );
      if ( p.getDistanceTo( opppos ) <= oppattrdist ) {
	px += opppos.getX() * oppattrx;
	py += opppos.getY() * oppattry;
      }
    }
  }

  homepos.setVecPosition( px, py );
}

void CLEval::UpdatePosPosition( Region * r )
{
  double px, py;
  VecPosition mypos = wm->getAgentGlobalPosition();
  VecPosition ballpos, opppos;
  ObjectT opp;

  r->GetClosestPointInRegion( ballpos.getX(), ballpos.getY(), &px, &py );
  VecPosition p( px, py );

  if ( p.getDistanceTo( ballpos ) <= ballattrdist ) {
    px += ballpos.getX() * ballattrx;
    py += ballpos.getY() * ballattry;
  }

  for ( int i = 0; i < 11; i++ ) {
    opp = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( wm->getConfidence( opp ) > ps->getOtherPlayerConfThr() ) {
      opppos = wm->getGlobalPosition( opp );
      if ( p.getDistanceTo( opppos ) <= oppattrdist ) {
	px += opppos.getX() * oppattrx;
	py += opppos.getY() * oppattry;
      }
    }
  }

  pospos.setVecPosition( px, py );
}

void CLEval::GetMarkInfo( VecPosition * pos, ObjectT * m )
{
  *pos = markpos;
  *m = mark;
}

bool CLEval::ClearUnadvised( double tx, double ty )
{
  int mynum = wm->getPlayerNumber();

  RBlist *dl = rulebase->GetActiveDirecs();
  while ( dl ) {

    if ( dl->dir->GetUnum()->IsSet( mynum )
	 && dl->dir->GetMode() == false ) {
      actionList *al = dl->dir->GetActions();
      while ( al ) {
	if ( al->act->GetType() == CLEAR ) {
	  if ( al->act->GetRegionArg()->PointInRegion( tx, ty ) ) {
	    Log.log( 653, "Clear to point %.2f, %.2f is unadvised", tx, ty );
	    return true;
	  }
	}
	al = al->next;
      }
    }
    dl = dl->next;
  }
  return false;

}

bool CLEval::DribbleUnadvised( double tx, double ty )
{
  int mynum = wm->getPlayerNumber();

  RBlist *dl = rulebase->GetActiveDirecs();
  while ( dl ) {

    if ( dl->dir->GetUnum()->IsSet( mynum )
	 && dl->dir->GetMode() == false ) {
      actionList *al = dl->dir->GetActions();
      while ( al ) {
	if ( al->act->GetType() == DRIBBLE ) {
	  if ( al->act->GetRegionArg()->PointInRegion( tx, ty ) )
	    return true;
	}
	al = al->next;
      }
    }
    dl = dl->next;
  }
  return false;
}

void CLEval::ChooseMarks()
{
  if ( wm->isBallKickable() ) {
    markeval = 0;
    return;
  }

  for ( int i = 0; i < 11; i++ ) {
    for ( int j = 0; j < 11; j++ ) {
      markmatrix[i][j] = false;
      marklmatrix[i][j] = false;
    }
  }

  RBlist *dl = rulebase->GetActiveDirecs();
  while ( dl ) {
    UpdateMarkAdvice( dl->dir );
    dl = dl->next;
  }

  int idx = wm->getPlayerNumber() - 1;
  int markp = -1;
  double bestdist = 10000;
  for ( int i = 0; i < 11; i++ )	// Treat (markl) with priority
  {
    if ( marklmatrix[idx][i] == true ) {
      Log.log( 705, "Maybe I'll markl #%d?", i );
      ObjectT mp = SoccerTypes::getOpponentObjectFromIndex( i );
      if ( wm->getConfidence( mp ) > ps->getOtherPlayerConfThr() ) {
	VecPosition opos = wm->getGlobalPosition( mp );
	double dist = opos.getDistanceTo( wm->getAgentGlobalPosition() );
	for ( int j = 0; j < 11; j++ ) {
	  if ( j != idx && marklmatrix[j][i] ) {

	    ObjectT tp = SoccerTypes::getTeammateObjectFromIndex( j );
	    if ( wm->getConfidence( tp ) > ps->getOtherPlayerConfThr() ) {
	      double dist2 =
		wm->getGlobalPosition( tp ).getDistanceTo( opos );
	      Log.log( 705,
		       "Player #%d is also marking %d.. his distance: %.3f... mine: %.3f",
		       j + 1, i + 1, dist2, dist );
	      if ( dist2 < dist ) {
		dist = 100000;
	      }
	    }
	  }
	}

	if ( dist < bestdist ) {
	  //bestdist = dist;
	  markp = i;
	  Log.log( 705, "So far, my mark is %d", markp + 1 );
	}
      }
    }
  }

  if ( markp != -1 ) {
    ObjectT opp = SoccerTypes::getOpponentObjectFromIndex( markp );
    markpos =
      ( wm->getGlobalPosition( opp ) * .7 ) +
      ( VecPosition( -52.5, 0 ) * .3 );
    Log.log( 705, "Decided to markl player #%d, at Position (%.2f, %.2f)",
	     markp + 1, markpos.getX(), markpos.getY() );
    markeval = interceptfastest * .9;
    return;

  }

  for ( int i = 0; i < 11; i++ )	// Treat (markl) with priority
  {
    if ( markmatrix[idx][i] == true ) {
      Log.log( 705, "Maybe I'll mark #%d?", i );
      ObjectT mp = SoccerTypes::getOpponentObjectFromIndex( i );
      if ( wm->getConfidence( mp ) > ps->getOtherPlayerConfThr() ) {
	VecPosition opos = wm->getGlobalPosition( mp );
	double dist = opos.getDistanceTo( wm->getAgentGlobalPosition() );
	for ( int j = 0; j < 11; j++ ) {
	  if ( j != idx && markmatrix[j][i] ) {

	    ObjectT tp = SoccerTypes::getTeammateObjectFromIndex( j );
	    if ( wm->getConfidence( tp ) > ps->getOtherPlayerConfThr() ) {
	      double dist2 =
		wm->getGlobalPosition( tp ).getDistanceTo( opos );
	      Log.log( 705,
		       "Player #%d is also marking %d.. his distance: %.3f... mine: %.3f",
		       j + 1, i + 1, dist2, dist );
	      if ( dist2 < dist ) {
		dist = 100000;
	      }
	    }

	  }
	}

	if ( dist < bestdist ) {
	  //bestdist = dist;
	  markp = i;
	  Log.log( 705, "So far, my mark is %d", markp + 1 );
	}
      }
    }
  }

  if ( markp != -1 ) {
    ObjectT opp = SoccerTypes::getOpponentObjectFromIndex( markp );
    markpos =
      ( wm->getGlobalPosition( opp ) * .7 ) + ( wm->getBallPos() * .3 );
    Log.log( 705, "Decided to markl player #%d, at Position (%.2f, %.2f)",
	     markp + 1, markpos.getX(), markpos.getY() );
    markeval = interceptfastest * .9;

  }
  markeval = 0;

}

VecPosition CLEval::GetPassReceivePoint()
{
  return passreceivept;
}

void CLEval::SetPassReceivePoint( VecPosition pt )
{
  passtoreceive = true;
  passreceivept = pt;
}
