/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _CLANG_REGION_
#define _CLANG_REGION_

#include <iostream>
#include "ClangTypes.h"

class Rule;

/// Region is the base case for all CLang region constructs
class Region
{
 public:
  /// All Region objects have a virtual deconstructor to handle memory management issues
  virtual ~ Region() {}

  /// Tests a point to see if it falls in a region
  virtual bool PointInRegion( double x, double y )
  {
    return false;
  }
  /// Output a region in its CLang parseable form
  virtual void Output( ostream & os, bool dv = true ) {}

  /// Returns true if the region can be used as a point
  virtual bool IsPoint()
  {
    return false;
  }
  virtual double Area()
  {
    return 0;
  }
  virtual void SetAssocRule( Rule * r )
  {
  }
  /// Gets the point in the region that is closest to (x,y) and stores it in (rx,ry)
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry )
  {
  }
} ;

/// regionList defines a linked list of regions to assist the UnionRegion object
struct regionList
{
  regionList()
  {
    reg = NULL;
    next = NULL;
  }

  Region *reg;
  regionList *next;
} ;

/// A PointRegion is a special type of region that is used to define more complex regions such as rectangles and triangles
class PointRegion: public Region
{
 public:
  PointRegion();
  virtual ~ PointRegion() {}

  /// Creates a PointRegion representing (pt x y)
  PointRegion( double x, double y );
  /// Creates a PointRegionr representing (pt team plr) or (pt ball) if plr = 0
  PointRegion( CLTeam t, int plr );
  virtual bool PointInRegion( double x, double y );

  virtual void Output( ostream & os, bool dv = true );

  /// Returns the X coordinate of the point
  virtual double GetX();
  /// Returns the Y coordinate of the point
  virtual double GetY();
  /// The area of a point is always 0
  virtual double Area()
  {
    return 0;
  }
  /// All subclasses of PointRegion are points
  virtual bool IsPoint()
  {
    return true;
  }
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
 private:
  bool usePlr;
  double x, y;
  CLTeam t;
  int plr;
} ;

/// RectRegion is a class representing a rectangular region
class RectRegion: public Region
{
 public:
  RectRegion();
  virtual ~RectRegion();
  /// Creates a region representing (rec topleft bottomright)
  RectRegion( PointRegion * topleft, PointRegion * bottomright );
  virtual bool PointInRegion( double x, double y );
  virtual void Output( ostream & os, bool dv = true );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
  virtual double Area();
private:
  PointRegion *tl;
  PointRegion *br;
} ;

/// TriRegion is a class representing a triangular region
class TriRegion: public Region
{
 public:
  TriRegion();
  virtual ~TriRegion();
  /// Creates a region representing (tri pt1 pt2 pt3)
  TriRegion( PointRegion * pt1, PointRegion * pt2, PointRegion * pt3 );
  virtual bool PointInRegion( double x, double y );
  virtual void Output( ostream & os, bool dv = true );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );

  virtual double Area();
 private:
  PointRegion *pt1;
  PointRegion *pt2;
  PointRegion *pt3;
} ;

/// ArcRegion is a class representing a region defined by two arcs
class ArcRegion: public Region
{
 public:
  ArcRegion();
  virtual ~ArcRegion();
  /// Creates a region corresponding to (arc c smrad lgrad begang spanang)
  ArcRegion( PointRegion * center, double smrad, double lgrad, double begang,
	     double spanang );
  virtual bool PointInRegion( double x, double y );
  virtual void Output( ostream & os, bool dv = true );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
  virtual double Area();

 private:
  PointRegion * c;
  double smrad, lgrad, begang, spanang;
} ;

/// NullRegion represents an empty region
class NullRegion: public Region
{
 public:
  /// Creates a region representing to (null)
  NullRegion() {}
  virtual ~NullRegion() {}
  virtual bool PointInRegion( double x, double y )
  {
    return false;
  }
  virtual void Output( ostream & os, bool dv = true );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
  virtual double Area()
  {
    return 0;
  }
} ;

/// UnionRegion represents a union of an arbitrary number of other regions of any type
class UnionRegion: public Region
{
 public:
  UnionRegion()
  {
    reglist = NULL;
  }
  virtual ~UnionRegion();
  /// Creates a UnionRegion representing the union of all the regions in rlist
  UnionRegion( regionList * rlist );
  virtual bool PointInRegion( double x, double y );
  virtual void Output( ostream & os, bool dv = true );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
  virtual double Area();
 private:
  regionList * reglist;
} ;

enum PtArithOp { 
  PTADD, PTSUBTRACT, PTMULTIPLY, PTDIVIDE 
} ;

/// PointArithRegion represents a PointRegion obtained by simple arithmetic operations on two other PointRegions
class PointArithRegion: public PointRegion
{
 public:
  PointArithRegion();
  virtual ~ PointArithRegion();
  /// Creates a PointArithRegion representing (pt1 op pt2)
  PointArithRegion( PointRegion * pt1, PointRegion * pt2, PtArithOp op );
  virtual double GetX();
  virtual double GetY();
  virtual void Output( ostream & os, bool dv = true );
 private:
  PointRegion * pt1;
  PointRegion *pt2;
  PtArithOp op;
} ;

/// A VarPtRegion is a special kind of point region that is represented by a CLang String and resolved when it is called
class VarPtRegion: public PointRegion
{
 public:
  VarPtRegion( CLangVar * v )
  {
    this->v = v;
  }
  virtual ~ VarPtRegion()
  {
    delete v;
  }
  virtual void Output( ostream & os, bool dv = true );
  virtual double GetX();
  virtual double GetY();
  virtual void SetAssocRule( Rule * r );
 private:
  CLangVar * v;
} ;

/// A UnumSetPoint represents a construct of the form (pt <team> <clangvar>)
class UnumSetPoint: public PointRegion
{
 public:
  UnumSetPoint()
  {
    v = NULL;
  }
   ~UnumSetPoint()
  {
    delete v;
  }
  UnumSetPoint( CLTeam t, CLangVar * v ) {
    this->t = t;
    this->v = v;
  }
  virtual void SetAssocRule( Rule * r )
  {
    this->r = r;
  }
  virtual double GetX();
  virtual double GetY();
  virtual void Output( ostream & os, bool dv = true );
 private:
  CLTeam t;
  CLangVar *v;
  Rule *r;
} ;

/// A VarRegion is a special kind of region that is represented by a CLang String and resolved when it is called
class VarRegion: public Region
{
public:
  VarRegion( CLangVar * v )
  {
    this->v = v;
  }
  virtual ~ VarRegion()
  {
    delete v;
  }
  virtual void Output( ostream & os, bool dv = true );
  virtual bool PointInRegion( double x, double y );
  virtual void GetClosestPointInRegion( double x, double y, double *rx,
					double *ry );
  virtual void SetAssocRule( Rule * r );
  virtual double Area();

private:
  CLangVar * v;
} ;

#endif
