/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include "ClangRule.h"
#include "ClangTypes.h"
#include "ClangDirective.h"
#include "ClangRuleBase.h"

ruleList::ruleList()
{
  r = NULL;
  next = NULL;
}

ruleList::~ruleList()
{
  delete r;
}

void Rule::InstantiateVariable( UnumSet * u )
{
  if ( !u->ContainsVariable() )
    return;
  USlist *nl = new USlist;
  nl->s = u;
  nl->next = inst;
  inst = nl;
}

void Rule::ClearInstantiations()
{
  USlist *c = inst, *t;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }

  inst = NULL;
}

UnumSet *Rule::GetInstantiatedVariable( CLangVar * v )
{
  USlist *c = inst;
  while ( c ) {
    if ( c->s->GetCLangVar() == NULL ) {
      cerr << "Unumset with no variable instantiated ??? ";
      c->s->Output( cerr );
      cerr << endl;
    }
    else if ( *( c->s->GetCLangVar() ) == *v ) {
      return c->s;
    }
    c = c->next;
  }

  //cerr << "Warning: Attempted to access uninstantiated variable ";
  //v->Output(cerr);
  //cerr  << endl;
  return NULL;

}

/// Stuff for CondDirRule

CondDirRule::CondDirRule( Condition * c, directiveList * d )
{
  this->c = c;
  this->d = d;
  used = false;
  Rule::Rule();
}

CondDirRule::~CondDirRule()
{
  delete c;
  directiveList *cur = d, *tmp;;

  while ( cur ) {
    tmp = cur;
    cur = cur->next;
    delete tmp;
  }
  ClearInstantiations();
}

void CondDirRule::Reset()
{
  used = false;
}

void CondDirRule::AddDirectivesToRulebase( bool direcm )
{
  if ( used )
    return;

  used = true;
  ClearInstantiations();
  if ( !c->Evaluate( this ) ) {
    return;
  }

  directiveList *cur = d;
  while ( cur ) {
    if ( direcm ) {
      if ( cur->dir == NULL )
	cout << "MAJOR ERROR! NULL DIRECTIVE!" << endl;
      cur->dir->SetAssocRule( this );
      rulebase->AddToActiveDirecs( cur->dir );
    }
    else {
      cur->dir->SetAssocRule( this );
      rulebase->AddToActiveModels( cur->dir );
    }
    cur = cur->next;
  }
}

void CondDirRule::Output( ostream & os, bool dispvar )
{
  os << "(";
  c->Output( os, dispvar );
  os << " ";
  directiveList *cur = d;
  while ( cur ) {
    cur->dir->Output( os, dispvar );
    os << " ";
    cur = cur->next;
  }
  os << ")";
}

/// CondRuleRule Stuff

CondRuleRule::CondRuleRule( Condition * c, ruleList * l )
{
  this->c = c;
  this->l = l;
  used = false;
  Rule::Rule();
}

CondRuleRule::~CondRuleRule()
{
  delete c;
  ruleList *cur = l, *tmp;
  while ( cur ) {
    tmp = cur;
    cur = cur->next;
    delete tmp;
  }
  ClearInstantiations();
}

void CondRuleRule::Reset()
{
  used = false;
  ruleList *cur = l;

  while ( cur ) {
    cur->r->Reset();
    cur = cur->next;
  }
}

void CondRuleRule::AddDirectivesToRulebase( bool direcm )
{
  if ( used )
    return;
  used = true;
  ClearInstantiations();
  if ( !c->Evaluate( this ) )
    return;

  ruleList *cur = l;

  while ( cur ) {
    cur->r->Reset();
    cur->r->AddDirectivesToRulebase( direcm );
    cur = cur->next;
  }

}

void CondRuleRule::Output( ostream & os, bool dispvar )
{
  os << "(";
  c->Output( os, dispvar );
  os << " ";
  ruleList *cur = l;
  while ( cur ) {
    cur->r->Output( os, dispvar );
    os << " ";
    cur = cur->next;
  }
  os << ")";
}

/// Stuff for IdListRule

IDListRule::IDListRule( idList * l )
{
  this->l = l;
  used = false;
  Rule::Rule();
}

IDListRule::~IDListRule()
{
  idList *cur = l, *tmp;
  while ( cur ) {
    tmp = cur;
    cur = cur->next;
    delete tmp;
  }
  ClearInstantiations();
}

void IDListRule::Reset()
{
  used = false;
}

void IDListRule::AddDirectivesToRulebase( bool direcm )
{
  if ( used )
    return;
  used = true;
  ClearInstantiations();

  idList *cur = l;

  while ( cur ) {
    Rule *rl = rulebase->ResolveRule( cur->r );
    if ( rl ) {
      rl->Reset();
      rl->AddDirectivesToRulebase( direcm );
    }
    cur = cur->next;
  }

}

void IDListRule::Output( ostream & os, bool dispvar )
{
  idList *cur = l;
  if ( cur->next == NULL ) {
    cur->r->Output( os );
    return;
  }

  os << "(";

  while ( cur ) {
    cur->r->Output( os );
    os << " ";
    cur = cur->next;
  }
  os << ")";
}
