/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Texas nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <iostream>
#include <fstream>
#include "ClangRuleBase.h"
#include "ClangRule.h"
#include "ClangParse.h"
#include "ClangRBInterface.h"

using namespace std;

RuleBase *rulebase;

void CLangVar::Set( char *a )
{
  strcpy( cvar, a );
}

CLangVar::CLangVar()
{
  Set( "" );
}

CLangVar::CLangVar( char *a )
{
  Set( a );
}

CLangVar & CLangVar::operator =( char *a )
{
  Set( a );
  return *this;
}

void CLangVar::Output( ostream & os )
{
  os << cvar;
}

bool CLangVar::operator ==( CLangVar & v )
{
  return ( strcmp( cvar, v.cvar ) == 0 );
}

// RuleBase stuff

RuleBase::~RuleBase()
{
  definemsg *c, *t;

  c = regs;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }

  c = conds;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }

  c = acts;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }

  c = dirs;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }

  c = rules;
  while ( c ) {
    t = c;
    c = c->next;
    delete t;
  }
  DeleteLists();
}

bool RuleBase::IsDefinedVar( CLangVar * v )
{
  return IsRule( v );
}

bool RuleBase::IsDefinedStr( CLangVar * v )
{
  return IsRegion( v ) || IsCondition( v ) || IsDirective( v )
    || IsAction( v );
}

bool RuleBase::IsUndefinedVar( CLangVar * v )
{
  return ( v->GetName()[0] != '\"' ) && ( !IsDefinedVar( v ) );
}

bool RuleBase::IsUndefinedStr( CLangVar * v )
{
  return ( v->GetName()[0] == '\"' ) && ( !IsDefinedStr( v ) );
}

bool RuleBase::IsRule( CLangVar * v )
{
  definemsg *c = rules;
  while ( c ) {
    if ( *( c->name ) == *v )
      return true;
    c = c->next;
  }
  return false;
}

bool RuleBase::IsRegion( CLangVar * v )
{
  definemsg *c = regs;
  while ( c ) {
    if ( *( c->name ) == *v )
      return true;
    c = c->next;
  }
  return false;
}

bool RuleBase::IsCondition( CLangVar * v )
{
  definemsg *c = conds;
  while ( c ) {
    if ( *( c->name ) == *v )
      return true;
    c = c->next;
  }
  return false;
}

bool RuleBase::IsDirective( CLangVar * v )
{
  definemsg *c = dirs;
  while ( c ) {
    if ( *( c->name ) == *v )
      return true;
    c = c->next;
  }
  return false;
}

bool RuleBase::IsAction( CLangVar * v )
{
  definemsg *c = acts;
  while ( c ) {
    if ( *( c->name ) == *v )
      return true;
    c = c->next;
  }
  return false;
}

bool RuleBase::IsPointRegion( CLangVar * v )
{
  definemsg *c = regs;
  while ( c ) {
    if ( *( c->name ) == *v )
      return c->reg->IsPoint();
    c = c->next;
  }
  return false;

}

void RuleBase::HearDefineMsg( definemsg * m )
{
  definemsg *cur, *l = NULL;

  if ( coachmode ) {
    UpdateMsg *msg = new UpdateMsg;
    msg->type = m->type;
    msg->var = new CLangVar( *m->name );
    msg->isrule = false;
    msg->next = NULL;
    msg->direcmode = m->rldirecmode;

    if ( defupdatel == NULL ) {
      defupdatel = defupdate = msg;
    }
    else {
      defupdatel->next = msg;
      defupdatel = msg;
    }

  }

  if ( m->type == DEFR ) {
    cur = regs;
    while ( cur ) {
      // Previously defined region, overwrite
      if ( m->name == cur->name ) {
	if ( l == NULL ) {
	  m->next = cur->next;
	  regs = m;
	  delete cur;
	}
	else {
	  m->next = cur->next;
	  l->next = m;
	  delete cur;
	}
	return;
      }
      l = cur;
      cur = cur->next;
    }
    if ( regs == NULL ) {
      regs = m;
      m->next = NULL;
    }
    else {
      m->next = regs;
      regs = m;
    }

  }
  else if ( m->type == DEFC ) {
    cur = conds;
    while ( cur ) {
      if ( m->name == cur->name ) {
	if ( l == NULL ) {
	  m->next = cur->next;
	  conds = m;
	  delete cur;
	}
	else {
	  m->next = cur->next;
	  l->next = m;
	  delete cur;
	}
	return;
      }
      l = cur;
      cur = cur->next;
    }
    if ( conds == NULL ) {
      conds = m;
      m->next = NULL;
    }
    else {
      m->next = conds;
      conds = m;
    }

  }
  else if ( m->type == DEFD ) {
    cur = dirs;
    while ( cur ) {
      if ( m->name == cur->name ) {
	if ( l == NULL ) {
	  m->next = cur->next;
	  dirs = m;
	  delete cur;
	}
	else {
	  m->next = cur->next;
	  l->next = m;
	  delete cur;
	}
	return;
      }
      l = cur;
      cur = cur->next;
    }
    if ( dirs == NULL ) {
      dirs = m;
      m->next = NULL;
    }
    else {
      m->next = dirs;
      dirs = m;
    }

  }
  else if ( m->type == DEFA ) {
    cur = acts;
    while ( cur ) {
      if ( m->name == cur->name ) {
	if ( l == NULL ) {
	  m->next = cur->next;
	  acts = m;
	  delete cur;
	}
	else {
	  m->next = cur->next;
	  l->next = m;
	  delete cur;
	}
	return;
      }
      l = cur;
      cur = cur->next;
    }
    if ( acts == NULL ) {
      acts = m;
      m->next = NULL;
    }
    else {
      m->next = acts;
      acts = m;
    }

  }
  else if ( m->type == DEFRL ) {
    cur = rules;
    while ( cur ) {
      if ( m->name == cur->name ) {
	m->rlon = cur->rlon;
	if ( l == NULL ) {
	  m->next = cur->next;
	  rules = m;
	  delete cur;
	}
	else {
	  m->next = cur->next;
	  l->next = m;
	  delete cur;
	}
	return;
      }
      l = cur;
      cur = cur->next;
    }
    if ( rules == NULL ) {
      rules = m;
      m->next = NULL;
    }
    else {
      m->next = rules;
      rules = m;
    }

  }

}

Region *RuleBase::ResolveRegion( CLangVar * v )
{
  definemsg *cur = regs;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->reg;
    cur = cur->next;
  }
  return NULL;
}

PointRegion *RuleBase::ResolvePtRegion( CLangVar * v )
{
  definemsg *cur = regs;
  while ( cur ) {

    if ( *( cur->name ) == *v ) {
      if ( cur->reg->IsPoint() )
	return ( PointRegion * ) cur->reg;
    }
    cur = cur->next;
  }
  return NULL;
}

Condition *RuleBase::ResolveCondition( CLangVar * v )
{
  definemsg *cur = conds;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->cond;
    cur = cur->next;
  }
  return NULL;
}

Directive *RuleBase::ResolveDirective( CLangVar * v )
{
  definemsg *cur = dirs;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->dir;
    cur = cur->next;
  }
  return NULL;
}

Action *RuleBase::ResolveAction( CLangVar * v )
{
  definemsg *cur = acts;
  while ( cur ) {

    if ( *( cur->name ) == *v ) {
      /*cout << "Action resolve" << endl;
         cur->act->Output(cout);
         cout << "EAR" << endl; */
      return cur->act;
    }
    cur = cur->next;
  }
  return NULL;
}

Rule *RuleBase::ResolveRule( CLangVar * v )
{
  definemsg *cur = rules;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->rl;
    cur = cur->next;
  }
  return NULL;
}

bool RuleBase::IsRuleOn( CLangVar * v )
{
  definemsg *cur = rules;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->rlon;
    cur = cur->next;
  }
  return false;
}

bool RuleBase::IsRuleOff( CLangVar * v )
{
  definemsg *cur = rules;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return !cur->rlon;
    cur = cur->next;
  }
  return false;
}

bool RuleBase::IsRuleModel( CLangVar * v )
{
  definemsg *cur = rules;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return !cur->rldirecmode;
    cur = cur->next;
  }
  return false;
}

bool RuleBase::IsRuleDirec( CLangVar * v )
{
  definemsg *cur = rules;
  while ( cur ) {

    if ( *( cur->name ) == *v )
      return cur->rldirecmode;
    cur = cur->next;
  }
  return false;
}

void RuleBase::ChangeRuleState( CLangVar * v, bool onoff )
{
  definemsg *cur = rules;
  while ( cur ) {
    if ( *( cur->name ) == *v ) {
      cur->rlon = onoff;
      if ( coachmode ) {
	UpdateMsg *m = new UpdateMsg;
	m->var = new CLangVar( *v );
	m->next = NULL;
	m->onoff = onoff;
	m->isrule = true;
	if ( ruleupdatel == NULL ) {
	  ruleupdate = ruleupdatel = m;
	}
	else {
	  ruleupdatel->next = m;
	  ruleupdatel = m;
	}

      }
      else {
	//      cout << "Changing rule state for ";
	//      v->Output(cout);
	//      cout << " to " << onoff << endl;
      }
    }
    cur = cur->next;
  }
}

void RuleBase::ChangeRuleState( bool onoff )
{
  definemsg *cur = rules;
  while ( cur ) {

    cur->rlon = onoff;
    cur = cur->next;
  }

  if ( coachmode ) {
    UpdateMsg *m = new UpdateMsg;
    m->var = new CLangVar( "all" );
    m->next = NULL;
    m->onoff = onoff;
    m->isrule = true;
    if ( ruleupdatel == NULL ) {
      ruleupdate = ruleupdatel = m;
    }
    else {
      ruleupdatel->next = m;
      ruleupdatel = m;
    }
  }

}

void RuleBase::ChangeRuleState( char *v, bool onoff )
{
  CLangVar var( v );
  ChangeRuleState( &var, onoff );
}

void RuleBase::ChangeRuleState( idList * r, bool onoff )
{
  idList *c = r;
  while ( c ) {
    ChangeRuleState( c->r, onoff );
    c = c->next;
  }
}

bool RuleBase::HearMsg( char *msg )
{
  parsestruct ctrl;
  ctrl.str = msg;
  ctrl.off = 0;
  return !yyparse( &ctrl );
}

void RuleBase::LoadFromFile( char *str, int *pmap, int *ourteam )
{
  ifstream f( str );
  if ( f )
    LoadFromFile( f, pmap, ourteam );
  else
    cerr << "Error opening file: " << str << endl;

  f.close();
}

void RuleBase::LoadFromFile( istream & f, int *mapping, int *ourteam )
{

  char *buf = new char[8500];
  char c;
  bool incomment = false;
  bool reading = false;
  bool sub = false;
  bool readinc = false;
  bool dollarsteamread = false;
  bool dollarsoneread = false;
  int bo = 0;
  int ifoff = 0;
  char subteam = 'T';
  char ifname[255];

  bool dollarsread = false;
  while ( f.read( &c, 1 ) ) {
    if ( c == '\t' )
      c = ' ';

    if ( c == '#' ) {
      incomment = true;
    }
    else if ( dollarsread && !reading && !readinc && c == 'I' ) {
      readinc = true;
    }
    else if ( readinc && c == '\n' ) {
      ifname[ifoff] = '\0';
      LoadFromFile( ifname, mapping, ourteam );
      ifoff = 0;
      readinc = false;
      dollarsread = false;
    }
    else if ( readinc && c != '\n' ) {
      ifname[ifoff++] = c;
    }
    else if ( c == '%' ) {
      if ( !incomment ) {
	reading = !reading;
	if ( !reading ) {
	  buf[bo] = '\0';
	  if ( sub ) {
	    cout << "Result of sub: " << buf << endl;
	    sub = false;
	  }
	  if ( !HearMsg( buf ) ) {
	    cout << "Coach couldn't understand " << buf << endl;

	    //return ;
	  }
	  bo = 0;
	}
      }
    }
    else if ( c == '\n' ) {
      if ( incomment )
	incomment = false;
      else
	buf[bo++] = ' ';

    }
    else if ( c == '$' ) {
      dollarsread = true;
      dollarsoneread = false;
      dollarsteamread = false;
      sub = true;
    }
    else {
      if ( reading && !incomment && !dollarsread )
	buf[bo++] = c;
      else if ( reading && !incomment && dollarsread && mapping ) {
	if ( ( c == 'T' || c == 'O' ) && !dollarsteamread ) {
	  dollarsteamread = true;
	  subteam = c;

	}
	else if ( c == '1' && dollarsteamread && !dollarsoneread ) {
	  dollarsoneread = true;
	}
	else if ( c == '1' && dollarsteamread && dollarsoneread ) {
	  int n = 0;
	  if ( subteam == 'T' ) {
	    n = mapping[10];
	  }
	  else {
	    n = mapping[21];
	  }
	  if ( n >= 10 )
	    buf[bo++] = '1';
	  buf[bo++] = ( n % 10 ) + '0';
	  dollarsread = false;
	}
	else if ( ( ( c >= '2' && c <= '9' ) || c == '0' ) && dollarsteamread
		  && dollarsoneread ) {
	  int n = 0;
	  if ( subteam == 'T' ) {
	    n = mapping[c - '0' + 9];
	  }
	  else {
	    n = mapping[11 + c - '0' + 9];
	  }

	  if ( n >= 10 )
	    buf[bo++] = '1';
	  buf[bo++] = ( n % 10 ) + '0';
	  dollarsread = false;

	}
	else if ( ( ( c >= '2' && c <= '9' ) || c == '0' ) && dollarsteamread
		  && !dollarsoneread ) {
	  int n = 0;
	  if ( subteam == 'T' ) {
	    n = mapping[c - '0' - 1];
	  }
	  else {
	    n = mapping[11 + c - '0' - 1];
	  }

	  if ( n >= 10 )
	    buf[bo++] = '1';
	  buf[bo++] = ( n % 10 ) + '0';
	  dollarsread = false;
	}
	else {
	  if ( dollarsoneread ) {
	    int n = 0;
	    if ( subteam == 'T' ) {
	      n = mapping[0];
	    }
	    else {
	      n = mapping[11];
	    }

	    if ( n >= 10 )
	      buf[bo++] = '1';
	    buf[bo++] = ( n % 10 ) + '0';
	    dollarsread = false;

	  }

	  buf[bo++] = c;

	}
      }

    }
  }

  delete[]buf;
}

void RuleBase::DeleteRule( CLangVar * v )
{
  definemsg *m = rules, *l = NULL;

  while ( m ) {
    if ( *( m->name ) == *v ) {
      if ( l ) {
	l->next = m->next;
	delete m;
	return;
      }
      else {
	rules = rules->next;
	delete m;
	return;
      }
    }

    l = m;
    m = m->next;
  }

}

void RuleBase::DeleteRule()
{
  definemsg *m = rules, *t;
  while ( m ) {
    t = m;
    m = m->next;
    delete t;
  }
  rules = NULL;
}

void RuleBase::DeleteRule( idList * v )
{
  idList *l = v;
  while ( l ) {
    DeleteRule( l->r );
    l = l->next;
  }
}

void RuleBase::DeleteLists()
{

  RBlist *cr, *tr;
  cr = activemodels;
  while ( cr ) {
    tr = cr;
    cr = cr->next;
    delete tr;
  }

  cr = activedirecs;
  while ( cr ) {
    tr = cr;
    cr = cr->next;
    delete tr;
  }

  activemodels = NULL;
  activedirecs = NULL;
}

void RuleBase::UpdateLists()
{
  DeleteLists();

  definemsg *cur = rules;
  while ( cur ) {
    if ( cur->rldirecmode && cur->rlon ) {

      cur->rl->Reset();
      cur->rl->AddDirectivesToRulebase( true );
    }
    cur = cur->next;
  }

  cur = rules;
  while ( cur ) {
    if ( !cur->rldirecmode && cur->rlon ) {
      cur->rl->Reset();
      cur->rl->AddDirectivesToRulebase( false );
    }
    cur = cur->next;
  }

}

void RuleBase::AddToActiveDirecs( Directive * d )
{
/*	if (RBWM->getPlayerNumber() == 4)
	{
		cout << "Adding directive (";
		if (d == NULL)
			cout << "NULL?";
		else

			d->Output(cout);
		cout << ")" << endl;
	}*/
  RBlist *nd = new RBlist;
  nd->dir = d;
  nd->next = activedirecs;
  activedirecs = nd;

}

void RuleBase::AddToActiveModels( Directive * d )
{
  RBlist *nd = new RBlist;
  nd->dir = d;
  nd->next = activemodels;
  activemodels = nd;

}

void RuleBase::SetCoachMode( bool mode )
{
  coachmode = mode;
}

#define MAX_CLANG_MSG_SIZE      3025

bool RuleBase::NeedsUpdate( int mode )
{
  if ( mode & UPDATE_DEFINES ) {
    if ( defupdate != NULL )
      return true;
  }

  if ( mode & UPDATE_RULES ) {
    if ( ruleupdate != NULL )
      return true;
  }
  return false;
}

CLangVar *RuleBase::GetRuleName( Rule * r )
{
  definemsg *cur = rules;
  while ( cur ) {
    if ( cur->rl == r )
      return cur->name;
    cur = cur->next;
  }
  return NULL;
}
