/*
Copyright (c) 2003-2004, Gregory Kuhlmann, Peter Stone, Justin Lallinger,
University of Texas at Austin
All right reserved

Based on:

Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Amsterdam nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*! \file CoachablePlayer.cpp

<pre>
<b>File:</b>          CoachablePlayer.cpp
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       03/03/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the definitions for the Player class,
               which is a superclass from BasicPlayer and contains the
               decision procedure to select the skills from the
               BasicPlayer.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/03/2001       Jelle Kok       Initial version created
</pre>
*/
#include <saymsglib/SayMsgEncoder.h>

#include "CoachablePlayer.h"
#include "Parse.h"
#ifndef WIN32
#include <sys/poll.h> // needed for 'poll'
#endif
#include <strstream>

extern Logger LogDraw;

/*!This is the constructor the CoachablePlayer class and calls the constructor of the
   superclass BasicPlayer.
   \param act ActHandler to which the actions can be sent
   \param wm WorldModel which information is used to determine action
   \param ss ServerSettings that contain parameters used by the server
   \param ps PlayerSettings that contain parameters important for the client
   \param strTeamName team name of this player
   \param dVersion version this basicplayer corresponds to
   \param iReconnect integer that defines player number (-1 when new player) */
CoachablePlayer::CoachablePlayer( ActHandler* act, WorldModel *wm, ServerSettings *ss,
      PlayerSettings *ps,
      CLEval *cl, int iNr, char* strTeamName, double dVersion, int iReconnect )

{
  char str[MAX_MSG];

  ACT           = act;
  WM            = wm;
  SS            = ss;
  PS            = ps;
  evaluator     = cl;
  bContLoop     = true;
  m_iPenaltyNr  = -1;
  WM->setTeamName( strTeamName );
  m_timeLastSay = -5;
  m_objMarkOpp  = OBJECT_ILLEGAL;
  m_actionPrev  = ACT_ILLEGAL;

  // wait longer as role number increases, to make sure players appear at the
  // field in the correct order
#ifdef WIN32
  Sleep( iNr*0 );
#else
  poll( 0, 0, iNr*0 );
#endif

  // create initialisation string
  if( iReconnect != -1 )
    sprintf( str, "(reconnect %s %d)", strTeamName, iReconnect );
  else if( iNr == 1 )
    sprintf( str, "(init %s (version %f) (goalie))", strTeamName, dVersion );
  else
    sprintf( str, "(init %s (version %f))", strTeamName, dVersion );
  ACT->sendMessage( str );
  
}

/*! This is the main loop of the agent. This method calls the update methods
    of the world model after it is indicated that new information has arrived.
    After this, the correct main loop of the player type is called, which
    puts the best soccer command in the queue of the ActHandler. */
void CoachablePlayer::mainLoop( )
{
  char str[MAX_MSG];
  Timing timer;

  // wait for new information from the server
  // cannot say bContLoop=WM->wait... since bContLoop can be changed elsewhere
  if(  WM->waitForNewInformation() == false )
    bContLoop =  false;

  // and set the clang version
  sprintf( str, "(clang (ver 8 8))" );
  ACT->sendMessage( str );

  Time lastRuleBaseUpdate = WM->getCurrentTime();
  while( bContLoop )                                 // as long as server alive
  {
    Log.logWithTime( 3, "  start update_all" );
    Log.setHeader( WM->getCurrentCycle() );
    LogDraw.setHeader( WM->getCurrentCycle() );

    if( WM->updateAll( ) == true )
    {
      timer.restartTime();

      // Update the rulebases list of active rules
      if ( lastRuleBaseUpdate != WM->getCurrentTime() ) {
	// Make sure the world model updates its idea of who is ball owner
	WM->getBallOwner();	
	rulebase->UpdateLists();
	lastRuleBaseUpdate = WM->getCurrentTime();
      }

      // Choose action
      SoccerCommand soc;
      if ( WM->getPlayerNumber() == 1 )
	soc = coachable_goalie();
      else
	soc = coachable();

      if( shallISaySomething(soc) == true ) {         // shall I communicate
	m_timeLastSay = WM->getCurrentTime();
	char strMsg[MAX_SAY_MSG];
	if( WM->getPlayerNumber() == 6 &&
	    WM->getBallPos().getX() < - PENALTY_X + 4.0  )
	  sayOppAttackerStatus( strMsg );
	else
	  sayBallStatus( strMsg );
	if( strlen( strMsg ) != 0 )
	  Log.log( 600, "send communication string: %s", strMsg );
	WM->setCommunicationString( strMsg );
      }
	
      Log.logWithTime( 3, "  determined action; waiting for new info" );
      // directly after see message, will nog get better info, so send commands
      if( WM->getTimeLastSeeMessage() == WM->getCurrentTime() ||
          (SS->getSynchMode() == true && WM->getRecvThink() == true ))
      {
        Log.logWithTime( 3, "  send messages directly" );
        ACT->sendCommands( );
        Log.logWithTime( 3, "  sent messages directly" );
        if( SS->getSynchMode() == true  )
        {
          WM->processRecvThink( false );
          ACT->sendMessageDirect( "(done)" );
        }
      }

    }
    else
      Log.logWithTime( 3, "  HOLE no action determined; waiting for new info");

    if( WM->getCurrentCycle()%(SS->getHalfTime()*SS->getSimulatorStep()) != 0 )
    {
      if( LogDraw.isInLogLevel( 600 ) )
      {
        WM->logDrawInfo( 600 );
      }

      if( LogDraw.isInLogLevel( 601 ) )
        WM->logDrawBallInfo( 601 );

      if( LogDraw.isInLogLevel( 700 ) )
        WM->logCoordInfo( 700 );
    }
 
    Log.logWithTime( 604, "time for action: %f", timer.getElapsedTime()*1000 );
           
    // wait for new information from the server cannot say
    // bContLoop=WM->wait... since bContLoop can be changed elsewhere
    if(  WM->waitForNewInformation() == false )
        bContLoop =  false;
  }

  // shutdow, print hole and number of players seen statistics
  printf("Shutting down player %d\n", WM->getPlayerNumber() );
  printf("   Number of holes: %d (%f)\n", WM->iNrHoles,
                         ((double)WM->iNrHoles/WM->getCurrentCycle())*100 );
  printf("   Teammates seen: %d (%f)\n", WM->iNrTeammatesSeen,
                         ((double)WM->iNrTeammatesSeen/WM->getCurrentCycle()));
  printf("   Opponents seen: %d (%f)\n", WM->iNrOpponentsSeen,
                         ((double)WM->iNrOpponentsSeen/WM->getCurrentCycle()));

}


/*! This method returns the position to move in case of a dead ball situation.
    A dead ball situation occurs when the team can have a free kick, kick in,
    etc. The agent will move to the position behind the ball and when he is
    there will move to the ball again. */
VecPosition CoachablePlayer::getDeadBallPosition(  )
{
  VecPosition pos, posBall = WM->getBallPos();
  VecPosition posAgent = WM->getAgentGlobalPosition();
  double dDist;

  // determine point to move to
  if( WM->isKickInUs()  )
    pos = posBall + VecPosition( -1.5, sign( posBall.getY() )*1.5 );
  else if( WM->isCornerKickUs( ) )
    pos = posBall + VecPosition( 1.5, sign( posBall.getY() ) * 1.5 );
  else if( WM->isFreeKickUs() || WM->isOffsideThem() || WM->isGoalKickUs() ||
           WM->isFreeKickFaultThem() || WM->isBackPassThem() )
    pos = posBall + VecPosition( -1.5, 0.0 );
  else
    return VecPosition( UnknownDoubleValue, UnknownDoubleValue );

  AngDeg      angBall = (posBall-posAgent).getDirection() ;
  ObjectT     obj = WM->getClosestInSetTo( OBJECT_SET_PLAYERS,
                                           WM->getAgentObjectType(), &dDist);
  VecPosition posPlayer = WM->getGlobalPosition( obj );

  // change point when heading towards other player or towards the ball
  if( fabs( angBall - (posPlayer-posAgent).getDirection() ) < 20 &&
      dDist < 6 )
    pos -= VecPosition( 5, 0 );
  if( fabs( angBall -  (pos-posAgent).getDirection()) < 20 )
  {
    angBall = VecPosition::normalizeAngle( angBall - 90 );
    pos = posBall + VecPosition( 1, angBall , POLAR );
  }
  return pos;
}

/********************** SAY **************************************************/

/*!This method determines whether a player should say something.
   \return bool indicating whether the agent should say a message */
bool CoachablePlayer::shallISaySomething( SoccerCommand socPri )
{
  bool        bReturn;

  bReturn  = ((WM->getCurrentTime() - m_timeLastSay) >= SS->getHearDecay());
  bReturn  &= amIAgentToSaySomething( socPri );
  bReturn  &= (WM->getCurrentCycle() > 0 );

  return bReturn;
}

/*! This method returns a boolean indicating whether I should communicate my
    world model to the other agents.
    \return boolean indicating whether I should communicate my world model. */
bool CoachablePlayer::amIAgentToSaySomething( SoccerCommand socPri )
{
  double  dDist;
  ObjectT obj;

  // get the closest teammate to the ball, if we are not him, we do not
  // communicate since he will
  obj = WM->getClosestInSetTo( OBJECT_SET_TEAMMATES,OBJECT_BALL,&dDist);
  if( dDist < SS->getVisibleDistance() &&
      obj != WM->getAgentObjectType() )
    return false;

  // in the defense, player 6 keeps track of the opponent attacker
  if( WM->getBallPos().getX() < - PENALTY_X + 4.0 &&
      WM->getConfidence( OBJECT_BALL ) > 0.96 &&
      WM->getPlayerNumber() == 6 &&
      WM->getCurrentCycle() % 3 == 0 ) // once very 3 cycles is enough
  {
    Log.log( 600, "player 6 is going to communicate attacker info" );
    return true;
  }

  VecPosition posBallPred;
  WM->predictBallInfoAfterCommand( socPri, &posBallPred );
  VecPosition posAgentPred = WM->predictAgentPosAfterCommand( socPri );
  // in all other cases inform teammates of ball when you have good information
  if( ( WM->getTimeChangeInformation(OBJECT_BALL) == WM->getCurrentTime() &&
          WM->getRelativeDistance( OBJECT_BALL ) < 20.0 &&
          WM->getTimeLastSeen( OBJECT_BALL ) == WM->getCurrentTime() )
        ||
      (
       WM->getRelativeDistance( OBJECT_BALL ) < SS->getVisibleDistance() &&
       WM->getTimeLastSeen( OBJECT_BALL ) == WM->getCurrentTime()  
       )
      ||
      ( // pass ball
       WM->getRelativeDistance( OBJECT_BALL ) < SS->getMaximalKickDist() &&
       posBallPred.getDistanceTo( posAgentPred ) > SS->getMaximalKickDist() 
       )
      )
    return true;

  return false;
}


void CoachablePlayer::sayOppAttackerStatus( char* strMsg )
{
  // find the closest opponent attacker to the penalty spot.
  double dDist ;
  ObjectT objOpp = WM->getClosestInSetTo( OBJECT_SET_OPPONENTS,
         VecPosition(- PITCH_LENGTH/2.0 + 11.0, 0 ), &dDist  ) ;

  if( objOpp == OBJECT_ILLEGAL || dDist >= 20.0 )
  {
    strncpy( strMsg, "", 0 );
    return;
  }

  VecPosition posOpp = WM->getGlobalPosition( objOpp );
  if( fabs( posOpp.getY() ) > 10 )
  {
    strncpy( strMsg, "", 0 );
    return;
  }

  SayMsgEncoder myencoder;
  myencoder.add( new OppPos( SoccerTypes::getIndex( objOpp ) + 1,
			     posOpp.getX(), posOpp.getY(), 1 ) );
  strcpy( strMsg, myencoder.getEncodedStr().c_str() );
  myencoder.clear();
}

/*! This method creates a string to communicate the ball status. When
    the player just kicks the ball, it is the new velocity of the
    ball, otherwise it is the current velocity.
    \param strMsg will be filled */
void CoachablePlayer::sayBallStatus( char * strMsg  )
{
  VecPosition posBall = WM->getGlobalPosition( OBJECT_BALL );
  VecPosition velBall = WM->getGlobalVelocity( OBJECT_BALL );
  int iDiff = 0;
  SoccerCommand soc = ACT->getPrimaryCommand();

  if( WM->getRelativeDistance( OBJECT_BALL ) < SS->getMaximalKickDist() )
  {
    // if kick and a pass
    if( soc.commandType == CMD_KICK )
    {
      WM->predictBallInfoAfterCommand( soc, &posBall, &velBall );
      VecPosition posAgent = WM->predictAgentPos( 1, 0 );
      if( posBall.getDistanceTo( posAgent ) > SS->getMaximalKickDist() + 0.2 )
        iDiff = 1;
    }

    if( iDiff == 0 )
    {
      posBall = WM->getGlobalPosition( OBJECT_BALL );
      velBall.setVecPosition( 0, 0 );
    }
  }
  Log.log( 600, "create comm. ball after: (%1.2f,%1.2f)(%1.2f,%1.2f) diff %d",
     posBall.getX(), posBall.getY(), velBall.getX(), velBall.getY(), iDiff);
  makeBallInfo( posBall, velBall, iDiff, strMsg );
}

void CoachablePlayer::makeBallInfo( VecPosition posBall, VecPosition velBall, int iDiff,
                            char * strMsg  )
{
  SayMsgEncoder myencoder;
  myencoder.add( new BallPos( posBall.getX(), posBall.getY(), 1 - iDiff ) ); 
  myencoder.add( new BallVel( velBall.getX(), velBall.getY(), 1 - iDiff ) ); 
  strcpy( strMsg, myencoder.getEncodedStr().c_str() );
  myencoder.clear();
}


SoccerCommand CoachablePlayer::coachable(  )
{
  return clang_act();
}

SoccerCommand CoachablePlayer::coachable_goalie(  )
{
  int i;
  SoccerCommand soc;
  VecPosition posAgent = WM->getAgentGlobalPosition();
  AngDeg angBody = WM->getAgentGlobalBodyAngle();

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -45, -PENALTY_AREA_WIDTH / 4.0 );
  static const VecPosition posRightTop( -45, +PENALTY_AREA_WIDTH / 4.0 );

  // define the borders of this rectangle using the two points.
  static Line lineFront =
    Line::makeLineFromTwoPoints( posLeftTop, posRightTop );
  static Line lineLeft =
    Line::makeLineFromTwoPoints( VecPosition( -50.0, posLeftTop.getY() ),
				 posLeftTop );
  static Line lineRight =
    Line::makeLineFromTwoPoints( VecPosition( -50.0, posRightTop.getY() ),
				 posRightTop );


  if ( WM->isBeforeKickOff() ) {
    ACT->putCommandInQueue( teleportToPos( VecPosition( -50, 0 ) ) );
    return soc;
  }

  if ( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() ) {	// confidence ball too  low
    ACT->putCommandInQueue( searchBall() );	// search ball
    ACT->putCommandInQueue( alignNeckWithBody() );
  }
  else if ( WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
	    WM->isCornerKickThem() ) {
    if ( WM->isBallCatchable() ) {
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else if ( WM->isBallKickable() ) {

      // Goalie Change here... calls clang_act() instead when ball is kickable,
      // but not catchable


      /*soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
         ACT->putCommandInQueue( soc );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) ); */

      soc = clang_act();
    }
    else if ( WM->isInOwnPenaltyArea( getInterceptionPointBall( &i, true ) )
	      && WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL,
					&i ) == WM->getAgentObjectType() ) {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else {
      // make line between own goal and the ball
      VecPosition posMyGoal = ( WM->getSide() == SIDE_LEFT )
	? SoccerTypes::getGlobalPositionFlag( OBJECT_GOAL_L, SIDE_LEFT )
	: SoccerTypes::getGlobalPositionFlag( OBJECT_GOAL_R, SIDE_RIGHT );
      Line lineBall =
	Line::makeLineFromTwoPoints( WM->getBallPos(), posMyGoal );

      // determine where your front line intersects with the line from ball
      VecPosition posIntersect = lineFront.getIntersection( lineBall );

      // outside rectangle, use line at side to get intersection
      if ( posIntersect.isRightOf( posRightTop ) )
	posIntersect = lineRight.getIntersection( lineBall );
      else if ( posIntersect.isLeftOf( posLeftTop ) )
	posIntersect = lineLeft.getIntersection( lineBall );

       VecPosition hpos;
         evaluator->Update();
         StockAction gh(STOCK_GO_HOME), gp(STOCK_GO_POS);
         if (evaluator->Evaluate(&gh) > evaluator->Evaluate(&gp))
         {
         hpos = evaluator->GetHomePos();
         }
         else
         {
         hpos = evaluator->GetPosPos();
         }

         posIntersect = hpos * .55 + posIntersect * .45; 
      if ( posIntersect.getX() < -49 )
	posIntersect.setX( -49 );

      // and move to this position
      if ( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) >
	   0.5 ) {
	int inttime = (int) ( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) / 0.7 + 0.4);
	soc =
	  moveToPos( posIntersect, PS->getPlayerWhenToTurnAngle(), 0, false,
		     inttime );
	ACT->putCommandInQueue( soc );
	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else {
	ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
  }
  else if ( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true ) {
    if ( WM->isBallKickable() ) {
      if ( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() ) {
	// move to position with lesser opponents.
	if ( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
				     Circle( posRightTop, 15.0 ) ) <
	     WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
				     Circle( posLeftTop, 15.0 ) ) )
	  soc.makeCommand( CMD_MOVE, posRightTop.getX(),
			   posRightTop.getY(), 0.0 );
	else
	  soc.makeCommand( CMD_MOVE, posLeftTop.getX(), posLeftTop.getY(),
			   0.0 );
	ACT->putCommandInQueue( soc );
      }
      else if ( WM->getTimeSinceLastCatch() > 28 ) {
	/*soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );
	   ACT->putCommandInQueue( soc ); */
	soc = clang_act();
      }
      else if ( WM->getTimeSinceLastCatch() < 25 ) {
	VecPosition posSide( 0.0, posAgent.getY() );
	if ( fabs( ( posSide - posAgent ).getDirection() - angBody ) > 10 ) {
	  soc = turnBodyToPoint( posSide );
	  ACT->putCommandInQueue( soc );
	}
	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if ( WM->isGoalKickUs() ) {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  else {
    ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }

  return soc;
}

SoccerCommand CoachablePlayer::clang_act()
{
  int mynum = WM->getPlayerNumber();
  ObjectT myObject = SoccerTypes::getTeammateObjectFromIndex( mynum - 1 );
  SoccerCommand soc( CMD_ILLEGAL );

  // If I don't know where I am or the ball is, I can't do much
  if ( WM->getConfidence( myObject ) < PS->getPlayerConfThr()
       || WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() ) {
    ACT->putCommandInQueue( searchBall() );
    ACT->putCommandInQueue( alignNeckWithBody() );
    Log.log( 650, "Don't know where the ball is... looking around" );
    return soc;
  }

  double eval = 0;
  evaluator->Update();

  // hack - may be masking error
  if ( tryToContinueLastAction && lastact != NULL ) {
    Log.log( 652, "Looking to see whether I can continue last action" );
    double eval = evaluator->Evaluate( lastact );
    if ( eval >= continueThr ) {
      strstream actstr;
      lastact->Output( actstr, false );
      actstr << ends;
      Log.log( 652, "Continuing to execute last action: %s", actstr.str() );
      soc = ExecuteCLangAction( lastact );
    }
    else {
      Log.log( 652, "The last action's was too low (less than %.3f)",
	       continueThr );
    }
  }
  if ( eval < continueThr ) {
    Action *bestact = evaluator->GetBestAction();
    if ( bestact ) {
      strstream actstr;
      bestact->Output( actstr, false );
      actstr << ends;
      Log.log( 652, "Executing best action: %s", actstr.str() );
      if ( bestact->GetType() != STOCK_GO_HOME &&
	   bestact->GetType() != STOCK_GO_POS &&
	   bestact->GetType() != STOCK_INTERCEPT &&
	   bestact->GetType() != STOCK_MARK )
	Log.log( 1000, "ACT %s", actstr.str() );
      soc = ExecuteCLangAction( bestact );
    }
    else {
      Log.log( 652, "I have no idea what to do!" );
    }
  }
  if ( soc.commandType != CMD_ILLEGAL )
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

  return soc;
}

SoccerCommand CoachablePlayer::ExecuteCLangAction( Action * a )
{
  SoccerCommand soc;
  Region *reg = NULL;
  //UnumSet *u = NULL;
  ObjectT opp;
  SideT myside = WM->getSide();
  SideT oppside = SIDE_RIGHT;
  if ( myside == SIDE_RIGHT )
    oppside = SIDE_LEFT;
  VecPosition pos;
  VecPosition mypos = WM->getAgentGlobalPosition();
  strstream ostr;
  double tx, ty, ang;
  int passto;
  tryToContinueLastAction = false;
  lastact = a;
  VecPosition dribvel;
  int lotime = 0, opptime;

  switch ( a->GetType() ) {
  case STOCK_GO_POS:
    pos = evaluator->GetPosPos();
    if ( pos.getX() > WM->getOffsideX() - .5 )
      pos.setX( WM->getOffsideX() - .5 );

    Log.log( 653,
	     "Going to position at (%.2f, %.2f) (can move? %d) (can teleport? %d)",
	     pos.getX(), pos.getY(), canMove(), canTeleport() );
    // Log.log(1000,"PTARG %.4f %.4f",pos.getX(),pos.getY());
//     Log.logPoint( 653, pos, ColorI( 255, 0, 0 ) );
//     Log.logLine( 653, mypos, pos, ColorI( 0, 0, 255 ) );
    if ( canTeleport() ) {
      if ( pos.getX() > -1 )
	pos.setX( -1 );
      soc = teleportToPos( pos );
      if ( pos.getDistanceTo( mypos ) < 3 ) {
	soc =
	  SoccerCommand( CMD_TURN,
			 90
			 /*VecPosition::normalizeAngle(WM->getAgentGlobalBodyAngle()+90) */
			  );
      }
    }
    else if ( canMove() ) {
      double mp = 1.0;
      if ( WM->getAgentTiredNess() == TIREDNESS_BAD )
	mp = 0.7;
      else if ( WM->getAgentTiredNess() == TIREDNESS_VERY_BAD )
	mp = 0.01;
      int inttime = (int) ( pos.getDistanceTo( mypos ) / mp / SS->getPlayerSpeedMax() + 0.45 );
      if ( pos.getDistanceTo( mypos ) > 3 )
	soc =
	  moveToPos( pos, PS->getPlayerWhenToTurnAngle(), 0, false, inttime );
      else
	soc = turnBodyToObject( OBJECT_BALL );
      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
	  1300 )
        soc.dPower = 30.0; // dash slow      
    }
    else {
      soc =
	SoccerCommand( CMD_TURN,
		       90
		       /*VecPosition::normalizeAngle(WM->getAgentGlobalBodyAngle()+90) */
		        );
    }
    break;


  case STOCK_GO_HOME:
    pos = evaluator->GetHomePos();
    if ( pos.getX() > WM->getOffsideX() - .5 )
      pos.setX( WM->getOffsideX() - .5 );
//     Log.logPoint( 653, pos, ColorI( 255, 0, 0 ) );
//     Log.logLine( 653, mypos, pos, ColorI( 0, 255, 0 ) );
    Log.log( 653,
	     "Going to home at (%.2f, %.2f) (can move? %d) (can teleport? %d)",
	     pos.getX(), pos.getY(), canMove(), canTeleport() );
    //Log.log(1000,"PTARG %.4f %.4f",pos.getX(),pos.getY());
    if ( canTeleport() ) {
      if ( pos.getX() > -1 )
	pos.setX( -1 );
      soc = teleportToPos( pos );
      Log.log( 653, "Teleporting to home position at (%.2f, %.2f)",
	       pos.getX(), pos.getY() );
    }
    else if ( canMove() ) {
      double mp = 1.0;
      if ( WM->getAgentTiredNess() == TIREDNESS_BAD )
	mp = 0.7;
      else if ( WM->getAgentTiredNess() == TIREDNESS_VERY_BAD )
	mp = 0.01;
      int inttime = (int) ( pos.getDistanceTo( pos ) / mp / SS->getPlayerSpeedMax() + 0.45 );
      if ( mypos.getDistanceTo( pos ) > 3 )
	soc =
	  moveToPos( pos, PS->getPlayerWhenToTurnAngle(), 0, false, inttime );
      else
	soc = turnBodyToObject( OBJECT_BALL );

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
	  SS->getRecoverDecThr()*SS->getStaminaMax()+500 )
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow      

      Log.log( 653, "Going to home position at (%.2f, %.2f)", pos.getX(),
	       pos.getY() );

    }
    else {
      soc =
	SoccerCommand( CMD_TURN,
		       /*VecPosition::normalizeAngle(WM->getAgentGlobalBodyAngle()+90) */
		       90 );
    }
    break;


  case STOCK_MARK:
    evaluator->GetMarkInfo( &pos, &opp );
    soc = moveToPos( pos, PS->getPlayerWhenToTurnAngle() );


    ACT->putCommandInQueue( turnNeckToObject( opp, soc ) );
    ACT->putCommandInQueue( soc );
    soc.commandType = CMD_ILLEGAL;
    break;




  case OLINE:
    // Not really sure how to "set an offside trap"
    Log.log( 653, "I'm supposed to set an offside trap" );
    reg = a->GetRegionArg();
//     Log.logRegion( 653, reg, ColorD( .5, 0, 0 ) );
    break;


  case HTYPE:
    // This is pretty much a useless action. I'm not even sure
    // if there's any real action to take, since you get a
    // message telling you when anyone's type is changed
    Log.log( 653, "My htype is now %d", a->GetIntArg() );
    break;


  case STOCK_PASS:
		passto = a->GetIntArg();

		{
		  if (evaluator->GetPassTargConf(passto) >= evaluator->minpassconf)
		    {
			VecPosition stp = evaluator->GetPassPoint(passto);
			double spd = evaluator->GetPassSpeed(passto);
                        int cycles =  evaluator->GetPassTime(passto);
//			soc = accelerateBallToVelocity(vel);
			Log.log(653,"Passing to %d at (%.3f, %.3f), conf: %f (min=%f)",
				passto, stp.getX(), stp.getY(),
				evaluator->GetPassTargConf(passto),
				evaluator->minpassconf);
                        Log.log(1000,"PTARG %.4f %.4f",stp.getX(),stp.getY());
                        double kspd = (spd * 10) * exp((double)cycles/SS->getBallDecay());
			soc = kickTo(stp, kspd);
			//Log.logPoint(699, stp, ColorI(255,0,255));
		    }
		  else
		    {
		      if (WM->getBallSpeed() > 0.1)
			{
			  Log.log(653,"Looking for player %d before pass, stopping ball, conf: %f (min=%f)",
				  passto, evaluator->GetPassTargConf(passto),
				  evaluator->minpassconf);
			  ACT->putCommandInQueue( freezeBall() );
			}
		      else
			{
			  if (WM->getCyclesHaveBall() >= 8)
			    {
			      Log.log(653,"Been looking for player %d before pass, conf: %f (min=%f), slow scanning",
				      passto, evaluator->GetPassTargConf(passto),
				      evaluator->minpassconf);
			      // If we haven't foudn for a while, just scan slowly.
			      ACT->putCommandInQueue( SoccerCommand( CMD_TURN, 45) );
			    }
			  else
			    {
			      Log.log(653,"Looking for player %d before pass, conf: %f (min=%f)",
				      passto, evaluator->GetPassTargConf(passto),
				      evaluator->minpassconf);
			      // the -1 is because those indices are 0 based
			      ACT->putCommandInQueue( turnBodyToObject(SoccerTypes::getTeammateObjectFromIndex(passto - 1)) );
			      ACT->putCommandInQueue( alignNeckWithBody() );
			    }
			}
		    }
		}



    break;
  case DRIBBLE:
  case STOCK_DRIBBLE_LEFT:
  case STOCK_DRIBBLE_RIGHT:
  case STOCK_DRIBBLE_CENTER:
    if ( a->GetType() == DRIBBLE ) {
      reg = a->GetRegionArg();
      reg->GetClosestPointInRegion( mypos.getX(), mypos.getY(), &tx,
				    &ty );
    }
    else if ( a->GetType() == STOCK_DRIBBLE_LEFT ) {
      tx = evaluator->stockdriblx;
      ty = evaluator->stockdribly;
    }
    else if ( a->GetType() == STOCK_DRIBBLE_RIGHT ) {
      tx = evaluator->stockdribrx;
      ty = evaluator->stockdribry;
    }
    else if ( a->GetType() == STOCK_DRIBBLE_CENTER ) {
      tx = evaluator->stockdribcx;
      ty = evaluator->stockdribcy;
    }


    dribvel =
      ( VecPosition( tx, ty ) - WM->getBallPos() ).normalize() * 2;

    for ( int i = 0; i < 11; i++ ) {
      opp = SoccerTypes::getOpponentObjectFromIndex( i );
      if ( WM->getConfidence( opp ) > PS->getOtherPlayerConfThr() ) {
	opptime = WM->CyclesToIntercept( opp, dribvel );
	if ( opptime < lotime )
	  lotime = opptime;
      }
    }




    pos.setVecPosition( tx, ty );
//     Log.logRegion( 653, reg, ColorD( .5, 0, 0 ) );
//     Log.logPoint( 653, VecPosition( tx, ty ), ColorD( 1.0, 0, 0 ) );
    if ( mypos.getDistanceTo( pos ) > 2 ) {
      Log.log( 1000, "PTARG %.4f %.4f", tx, ty );
      ang =
	atan2( ty - mypos.getY(), tx - mypos.getX() ) * 180 / 3.1415926;
      if ( lotime <= evaluator->fastdribblethr ) {
	Log.log( 653, "Dribbling fast to (%.2f, %.2f) at angle %.2f", tx, ty,
		 ang );
	soc = dribble( ang, DRIBBLE_FAST );
      }
      else if ( lotime <= evaluator->slowdribblethr ) {
	Log.log( 653, "Dribbling slow to (%.2f, %.2f) at angle %.2f", tx, ty,
		 ang );
	soc = dribble( ang, DRIBBLE_SLOW );
      }
      else {
	Log.log( 653, "Dribbling with ball to (%.2f, %.2f) at angle %.2f", tx,
		 ty, ang );
	soc = dribble( ang, DRIBBLE_WITHBALL );
      }
    }
    tryToContinueLastAction = true;
    continueThr = evaluator->Evaluate( a ) * .75;
    //soc = outplayOpponent(WM->getFastestInSetTo(OBJECT_SET_OPPONENTS,OBJECT_BALL), pos);
    break;


  case CLEAR:
  case STOCK_CLEAR_LEFT:
  case STOCK_CLEAR_RIGHT:
  case STOCK_CLEAR_CENTER:

    if ( a->GetType() == CLEAR ) {
      reg = a->GetRegionArg();
      reg->GetClosestPointInRegion( mypos.getX(), mypos.getY(), &tx,
				    &ty );
    }
    else if ( a->GetType() == STOCK_CLEAR_LEFT ) {
      tx = evaluator->stockclrlx;
      ty = evaluator->stockclrly;
    }
    else if ( a->GetType() == STOCK_CLEAR_RIGHT ) {
      tx = evaluator->stockclrrx;
      ty = evaluator->stockclrry;
    }
    else if ( a->GetType() == STOCK_CLEAR_CENTER ) {
      tx = evaluator->stockclrcx;
      ty = evaluator->stockclrcy;
    }
    Log.log( 653, "Clearing to (%.2f, %.2f)", tx, ty );
    Log.log( 1000, "PTARG %.4f %.4f", tx, ty );
//     Log.logRegion( 653, reg, ColorD( .5, 0, 0 ) );
//     Log.logPoint( 653, VecPosition( tx, ty ), ColorD( 1.0, 0, 0 ) );
    pos.setVecPosition( tx, ty );
    soc = kickTo( pos, SS->getBallSpeedMax() );
    break;


  case STOCK_SHOOT:

    pos = evaluator->GetShootPoint();

//     Log.logLine( 653, WM->getBallPos(), pos, ColorD( 1.0, 0, 0 ) );
    Log.log( 653, "Shooting to (%.2f, %.2f)", pos.getX(), pos.getY() );
    Log.log( 1000, "PTARG %.4f %.4f", pos.getX(), pos.getY() );
    /*if (WM->getBallSpeed() > .2)
       {
       tryToContinueLastAction = true;
       soc = freezeBall();
       }
       else */
    soc = kickTo( pos, SS->getBallSpeedMax() );
    break;


  case STOCK_HOLD:
    Log.log( 653, "Holding the ball" );
    if ( WM->getGlobalVelocity( OBJECT_BALL ).getMagnitude() > .1 )
      soc = freezeBall();
    else
      soc =
	SoccerCommand( CMD_TURN,
		       90
		       /*VecPosition::normalizeAngle(WM->getAgentGlobalBodyAngle()+90) */
		        );
    break;


  case STOCK_INTERCEPT:
    Log.log( 653, "Intercepting the ball" );
    if ( canMove() )		// Don't try to intercept the ball
      // if I can't move
      soc =
	intercept( WM->getPlayerNumber() ==
		   SoccerTypes::getIndex( WM->getOwnGoalieType() ) + 1 );
    else
      soc =
	SoccerCommand( CMD_TURN,
		       90
		       /*VecPosition::normalizeAngle(WM->getAgentGlobalBodyAngle()+90) */
		        );

//     Log.logPoint( 653, mypos, ColorI( 255, 0, 0 ) );
//     Log.logLine( 653, mypos, WM->getBallPos(), ColorI( 0, 0, 255 ) );
    break;


  case TACKLE:
    // This one will take some work
    // Tackle isn't one of the BasicPlayer's skills
    // Not even sure if uva knows about it?
    Log.log( 653, "I'd be tackling if it were implemented" );
    break;
  case STOCK_GOALIE_CATCH:
    Log.log( 653, "Catching the ball" );
    soc = catchBall();
  }
  if ( soc.commandType != CMD_ILLEGAL )
    ACT->putCommandInQueue( soc );
  return soc;
}


bool CoachablePlayer::canTeleport()
{
  return WM->isBeforeKickOff() || ( WM->getPlayMode() == PM_GOAL_LEFT )
    || ( WM->getPlayMode() == PM_GOAL_RIGHT );
}

ObjectT CoachablePlayer::getClosestOpponentInRegionTo( Region * r, ObjectT obj )
{
  VecPosition objpos = WM->getGlobalPosition( obj );
  VecPosition chkpos;
  double lowdist = 9999999;
  double dist = 0;
  ObjectT ret = OBJECT_OPPONENT_UNKNOWN;
  ObjectT check = OBJECT_OPPONENT_UNKNOWN;
  for ( int i = 0; i < 11; i++ ) {
    check = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( WM->getConfidence( check ) > PS->getOtherPlayerConfThr() ) {
      chkpos = WM->getGlobalPosition( check );
      if ( r->PointInRegion( chkpos.getX(), chkpos.getY() ) ) {
	dist = chkpos.getDistanceTo( objpos );
	if ( dist < lowdist ) {
	  ret = check;
	  lowdist = dist;
	}
      }
    }
  }

  return check;
}


ObjectT CoachablePlayer::getClosestOpponentInUnumSetTo( UnumSet * u, ObjectT obj )
{
  VecPosition objpos = WM->getGlobalPosition( obj );
  VecPosition chkpos;
  double lowdist = 9999999;
  double dist = 0;
  ObjectT ret = OBJECT_OPPONENT_UNKNOWN;
  ObjectT check = OBJECT_OPPONENT_UNKNOWN;
  for ( int i = 0; i < 11; i++ ) {
    check = SoccerTypes::getOpponentObjectFromIndex( i );
    if ( WM->getConfidence( check ) > PS->getPlayerConfThr() ) {
      chkpos = WM->getGlobalPosition( check );
      if ( u->IsSet( i + 1 ) ) {
	dist = chkpos.getDistanceTo( objpos );
	if ( dist < lowdist ) {
	  ret = check;
	  lowdist = dist;
	}
      }
    }
  }

  return check;
}

bool CoachablePlayer::canMove()
{

  if ( ( WM->getPlayMode() == PM_GOAL_KICK_LEFT
	 && WM->getSide() == SIDE_RIGHT )
       || ( WM->getPlayMode() == PM_GOAL_KICK_RIGHT
	    && WM->getSide() == SIDE_LEFT )
     )
    return false;
  else
    return true;
}
